import logging

from command.model.configuration import CMDRequestJson, CMDBooleanSchema, CMDStringSchema, CMDObjectSchema, \
    CMDArraySchema, CMDFloatSchema, CMDIntegerSchema
from schematics.models import Model
from schematics.types import StringType, BooleanType, ModelType, PolyModelType, BaseType
from swagger.utils import exceptions

from .fields import XmsClientNameField, XmsClientFlattenField, XmsClientDefaultField
from .fields import XmsClientRequestIdField, XNullableField, XPublishField, XRequiredField, XClientNameField, \
    XNewPatternField, XPreviousPatternField, XCommentField
from .fields import XmsParameterLocationField, XmsApiVersionField, XmsSkipUrlEncodingField, XmsArmIdDetailsField
from .fields import XmsSkipURLEncodingField, XAccessibilityField, XmsHeaderCollectionPrefix, XOriginalNameField
from .items import Items
from .reference import Reference, Linkable
from .schema import Schema, ReferenceSchema, schema_and_reference_schema_claim_function
from .x_ms_parameter_grouping import XmsParameterGroupingField

logger = logging.getLogger('aaz')


class ParameterBase(Model):
    """Describes a single operation parameter.
    A unique parameter is defined by a combination of a name and location.
    There are five possible parameter types.
    """

    IN_VALUE = None

    name = StringType(
        required=True)  # The name of the parameter. Parameter names are case sensitive. If in is "path", the name field MUST correspond to the associated path segment from the path field in the Paths Object. See Path Templating for further information. For all other cases, the name corresponds to the parameter name used based on the in property.
    description = StringType()  # A brief description of the parameter. This could contain examples of use.
    required = BooleanType(
        default=False)  # Determines whether this parameter is mandatory. If the parameter is in "path", this property is required and its value MUST be true. Otherwise, the property MAY be included and its default value is false.
    _in = StringType(
        serialized_name="in",
        deserialize_from="in",
        required=True
    )  # The location of the parameter. Possible values are "query", "header", "path", "formData" or "body".

    x_ms_parameter_grouping = XmsParameterGroupingField()  # TODO:
    x_ms_parameter_location = XmsParameterLocationField()  # TODO:
    x_ms_client_name = XmsClientNameField()  # TODO:
    x_ms_client_flatten = XmsClientFlattenField()  # TODO:
    x_ms_client_default = XmsClientDefaultField()
    x_ms_arm_id_details = XmsArmIdDetailsField()  # TODO: Add support for it, can be used for resource id template

    # specific properties
    _x_accessibility = XAccessibilityField()  # only used in ContainerRegistry Data plane
    _x_required = XRequiredField()  # only used in ContainerRegistry Data plane
    _x_publish = XPublishField()  # only used in Maps Data Plane
    _x_example = BaseType(serialized_name='x-example', deserialize_from='x-example')  # deprecated
    _x_examples = BaseType(serialized_name='x-examples', deserialize_from='x-examples')  # deprecated
    _x_client_name = XClientNameField()  # Only used in Maps Data Plane
    _x_new_pattern = XNewPatternField()  # Only used in FrontDoor Mgmt Plane
    _x_previous_pattern = XPreviousPatternField()  # Only used in FrontDoor Mgmt Plane
    _x_comment = XCommentField()  # Only used in IoTCenter Mgmt Plane
    _x_original_name = XOriginalNameField()  # Only used in Marketplane Catalog Data Plane

    @classmethod
    def _claim_polymorphic(cls, data):
        if isinstance(data, dict):
            in_value = data.get('in', None)
            return in_value is not None and in_value == cls.IN_VALUE
        elif isinstance(data, ParameterBase):
            return data.IN_VALUE == cls.IN_VALUE
        return False


class QueryParameter(Items, ParameterBase):
    """Parameters that are appended to the URL. For example, in /items?id=###, the query parameter is id."""
    IN_VALUE = "query"

    allow_empty_value = BooleanType(
        default=False,
        serialized_name="allowEmptyValue",
        deserialize_from="allowEmptyValue",
    )  # Sets the ability to pass empty-valued parameters. This is valid only for either query or formData parameters and allows you to send a parameter with a name only or an empty value. Default value is false.

    collection_format = StringType(
        choices=("csv", "ssv", "tsv", "pipes", "multi"),  # default is csv
        serialized_name="collectionFormat",
        deserialize_from="collectionFormat",
    )  # multi - corresponds to multiple parameter instances instead of multiple values for a single instance foo=bar&foo=baz.

    x_ms_api_version = XmsApiVersionField()
    x_ms_skip_url_encoding = XmsSkipUrlEncodingField()

    def to_cmd(self, builder, **kwargs):
        model = super().to_cmd(builder, **kwargs)

        model.name = self.name
        model.required = self.required
        builder.setup_description(model, self)

        if self.x_ms_skip_url_encoding:
            model.skip_url_encoding = False

        return model


class HeaderParameter(Items, ParameterBase):
    """Custom headers that are expected as part of the request."""
    IN_VALUE = "header"

    x_ms_client_request_id = XmsClientRequestIdField()

    # specific properties, will not support
    x_ms_header_collection_prefix = XmsHeaderCollectionPrefix()  # only used in Storage Data plane

    def _new_param(self):
        pass

    def to_cmd(self, builder, **kwargs):
        model = super().to_cmd(builder, **kwargs)
        model.name = self.name
        model.required = self.required

        builder.setup_description(model, self)
        return model


class PathParameter(Items, ParameterBase):
    """Used together with Path Templating, where the parameter value is actually part of the operation's URL. This does not include the host or base path of the API. For example, in /items/{itemId}, the path parameter is itemId."""
    IN_VALUE = "path"

    required = BooleanType(required=True, default=True)

    x_ms_skip_url_encoding = XmsSkipURLEncodingField()

    def to_cmd(self, builder, **kwargs):
        model = super().to_cmd(builder, **kwargs)
        model.name = self.name
        model.required = self.required

        builder.setup_description(model, self)

        if self.x_ms_skip_url_encoding:
            model.skip_url_encoding = True

        return model


class FormDataParameter(Items, ParameterBase):
    """Used to describe the payload of an HTTP request when either application/x-www-form-urlencoded, multipart/form-data or both are used as the content type of the request (in Swagger's definition, the consumes property of an operation). This is the only parameter type that can be used to send files, thus supporting the file type. Since form parameters are sent in the payload, they cannot be declared together with a body parameter for the same operation. Form parameters have a different format based on the content-type used (for further details, consult http://www.w3.org/TR/html401/interact/forms.html#h-17.13.4):
        - application/x-www-form-urlencoded - Similar to the format of Query parameters but as a payload. For example, foo=1&bar=swagger - both foo and bar are form parameters. This is normally used for simple parameters that are being transferred.
        - multipart/form-data - each parameter takes a section in the payload with an internal header. For example, for the header Content-Disposition: form-data; name="submit-name" the name of the parameter is submit-name. This type of form parameters is more commonly used for file transfers.
    """

    IN_VALUE = "formData"

    type = StringType(
        choices=("string", "number", "integer", "boolean", "array", "file"),
        required=True
    )  # If type is "file", the consumes MUST be either "multipart/form-data", " application/x-www-form-urlencoded" or both
    allow_empty_value = BooleanType(
        default=False,
        serialized_name="allowEmptyValue",
        deserialize_from="allowEmptyValue",
    )  # Sets the ability to pass empty-valued parameters. This is valid only for either query or formData parameters and allows you to send a parameter with a name only or an empty value. Default value is false.
    collection_format = StringType(
        choices=("csv", "ssv", "tsv", "pipes", "multi"),  # default is csv
        serialized_name="collectionFormat",
        deserialize_from="collectionFormat",
    )  # multi - corresponds to multiple parameter instances instead of multiple values for a single instance foo=bar&foo=baz.

    def to_cmd(self, builder, **kwargs):
        # TODO:
        raise exceptions.InvalidSwaggerValueError(
            msg="FormData Parameter is not supported",
            key=[self.name, self.type, self.format]
        )


class BodyParameter(ParameterBase, Linkable):
    """The payload that's appended to the HTTP request. Since there can only be one payload, there can only be one body parameter. The name of the body parameter has no effect on the parameter itself and is used for documentation purposes only. Since Form parameters are also in the payload, body and form parameters cannot exist together for the same operation."""

    IN_VALUE = "body"

    schema = PolyModelType(
        [ModelType(Schema), ModelType(ReferenceSchema)],
        claim_function=schema_and_reference_schema_claim_function,
        required=True
    )  # The schema defining the type used for the body parameter.

    x_nullable = XNullableField(
        default=False)  # TODO: # when true, specifies that null is a valid value for the associated schema

    def link(self, swagger_loader, *traces):
        if self.is_linked():
            return
        super().link(swagger_loader, *traces)

        self.schema.link(swagger_loader, *self.traces, 'schema')

    def to_cmd(self, builder, **kwargs):
        v = builder(self.schema, in_base=False, support_cls_schema=True)
        v.name = self.name
        if self.required:
            # A required body parameter cannot be frozen.
            # This can help to create an empty payload.
            v.frozen = False
        if v.frozen:
            logger.warning(
                msg=f"Request Body Parameter is None: {self.traces}"
            )
            return None
        if isinstance(v, (
                CMDStringSchema,
                CMDObjectSchema,
                CMDArraySchema,
                CMDBooleanSchema,
                CMDFloatSchema,
                CMDIntegerSchema
        )):
            model = CMDRequestJson()
            model.schema = v
            v.required = self.required
            if isinstance(v, CMDObjectSchema):
                # flatten body parameter
                v.client_flatten = True
        else:
            raise exceptions.InvalidSwaggerValueError(
                msg="Invalid Request type",
                key=self.traces,
                value=v.type
            )
        return model


class ParameterField(PolyModelType):

    def __init__(self, support_reference, **kwargs):
        model_spec = [
            QueryParameter, HeaderParameter, PathParameter, FormDataParameter, BodyParameter
        ]
        if support_reference:
            model_spec.append(Reference)
        super(ParameterField, self).__init__(model_spec=model_spec, **kwargs)



from schematics.models import Model
from schematics.types import StringType, URLType, PolyModelType
from .fields import ScopesField


class _SecuritySchemeBase(Model):
    """Allows the definition of a security scheme that can be used by the operations. Supported schemes are basic authentication, an API key (either as a header or as a query parameter) and OAuth2's common flows (implicit, password, application and access code)."""

    TYPE_VALUE = None

    type = StringType(choices=("basic", "apiKey", "oauth2"), required=True)
    description = StringType()

    @classmethod
    def _claim_polymorphic(cls, data):
        if isinstance(data, dict):
            type_value = data.get('type', None)
            return type_value is not None and type_value == cls.TYPE_VALUE
        elif isinstance(data, _SecuritySchemeBase):
            return data.TYPE_VALUE == cls.TYPE_VALUE
        return False


class BasicSecurityScheme(_SecuritySchemeBase):
    TYPE_VALUE = "basic"


class APIKeySecurityScheme(_SecuritySchemeBase):
    TYPE_VALUE = "apiKey"

    name = StringType(required=True)  # The name of the header or query parameter to be used.
    in_ = StringType(
        choices=("query", "header"),
        required=True,
        serialized_name="in",
        deserialize_from="in"
    )  # The location of the API key. Valid values are "query" or "header".


class OAuth2SecurityScheme(_SecuritySchemeBase):
    TYPE_VALUE = "oauth2"

    flow = StringType(choices=("implicit", "password", "application", "accessCode"), required=True)  # The flow used by the OAuth2 security scheme.
    authorization_url = URLType(
        serialized_name="authorizationUrl",
        deserialize_from="authorizationUrl"
    )  # The authorization URL to be used for this flow. oauth2 ("implicit", "accessCode")
    token_url = URLType(
        serialized_name="tokenUrl",
        deserialize_from="tokenUrl"
    )  # The token URL to be used for this flow. oauth2 ("password", "application", "accessCode")
    scopes = ScopesField()  # The available scopes for the OAuth2 security scheme.


class SecuritySchemeField(PolyModelType):

    def __init__(self, **kwargs):
        model_spec = [
            BasicSecurityScheme, APIKeySecurityScheme, OAuth2SecurityScheme
        ]
        super(SecuritySchemeField, self).__init__(model_spec=model_spec, **kwargs)




import datetime
import enum
import logging
import os
import re

from fuzzywuzzy import fuzz

from command.model.configuration import CMDResource
from ._utils import map_path_2_repo
from utils.base64 import b64encode_str
from swagger.utils import exceptions
from utils.case import to_singular

logger = logging.getLogger('aaz')


class Resource:
    _CAMEL_CASE_PATTERN = re.compile(r"^([a-zA-Z][a-z0-9]+)(([A-Z][a-z0-9]*)+)$")

    def __init__(self, resource_id, path, version, file_path, resource_provider, body):
        self.path = path
        self.id = resource_id
        self._version = ResourceVersion(version)
        self.file_path = file_path
        self.resource_provider = resource_provider
        self.file_path_version = self._get_file_path_version(file_path)

        operations = {}
        for method, v in body.items():
            if isinstance(v, dict) and 'operationId' in v:
                operations[v['operationId']] = method
        self.operations = operations

    @property
    def version(self):
        return self._version.version
    
    @property
    def rp_name(self):
        return self.resource_provider.name

    def __str__(self):
        return f"{self.path} {self.version}"

    def __hash__(self):
        return hash(f"{self.id} {self.version}")

    def get_operation_group_name(self):
        # TODO: Check database to fetch customized operation group name for resource_id
        if hasattr(self, "_op_group_name"):
            return self._op_group_name

        operation_groups = set()
        for operation_id, method in self.operations.items():
            op_group = self._parse_operation_group_name(operation_id, method)
            operation_groups.add(op_group)

        if None in operation_groups:
            return None

        if len(operation_groups) == 1:
            return operation_groups.pop()

        op_group_name = sorted(
            operation_groups,
            key=lambda nm: fuzz.partial_ratio(self.id, nm),  # use the name which is closest to resource_id
            reverse=True
        )[0]
        setattr(self, "_op_group_name", op_group_name)
        return op_group_name

    def _parse_operation_group_name(self, op_id, method):
        # extract operation group name from operation_id
        value = op_id.strip()
        value = value.replace('-', '_')
        if '_' in value:
            parts = value.split('_')
            op_group_name = parts[0]
            if op_group_name.lower() in ("create", "get", "update", "delete", "patch"):
                op_group_name = parts[1]
        else:
            if ' ' in value:
                value = value.replace(' ', '')  # Changed to Camel Case
            match = self._CAMEL_CASE_PATTERN.match(value)
            if not match:
                logger.error(f"InvalidOperationIdFormat:"
                             f"\toperationId should be in format of '[OperationGroupName]_[OperationName]' "
                             f"or '[Verb][OperationGroupName]':\n"
                             f"\tfile: {map_path_2_repo(self.file_path)}\n"
                             f"\tpath: {self.path}\n"
                             f"\tmethod: {method} operationId: {op_id}\n")
                return None
            op_group_name = match[2]  # [OperationGroupName]

        # Handle plural and singular cases
        words = []
        for part in self.id.split('?')[0].split('/'):
            if part == '{}' and len(words):
                singular = to_singular(words[-1])
                if singular:
                    words[-1] = singular
            else:
                words.append(part.replace('_', ""))
        op_group_singular = to_singular(op_group_name) or op_group_name
        words.reverse()  # search from tail
        for word in words:
            word_singular = to_singular(word) or word
            if len(word_singular) > 1 and op_group_singular.lower().endswith(word_singular.lower()):
                if word == word_singular:
                    # use singular
                    op_group_name = op_group_singular
                elif word != word_singular:
                    # use plural
                    op_group_plural = to_singular(op_group_singular)
                    if op_group_plural is not False:
                        op_group_name = op_group_plural
                break
        return op_group_name

    def _get_file_path_version(self, file_path):
        dir_parts = file_path.split(self.resource_provider.folder_path)[-1].split(os.sep)[:-1]
        dir_parts = [part for part in dir_parts if part]
        if len(dir_parts) < 2:
            raise exceptions.InvalidSwaggerValueError(f"Cannot parse file version", file_path)
        version = None
        for idx in range(len(dir_parts)):
            readiness = dir_parts[idx]
            if readiness.lower() in ('stable', 'preview') and idx + 1 < len(dir_parts):
                version = dir_parts[idx+1]
                break
        if not version:
            raise exceptions.InvalidSwaggerValueError(
                "Cannot parse version and readiness in file path", file_path)
        file_path_version = ResourceVersion(version)
        if file_path_version.readiness == ResourceVersion.Readiness.Stable and readiness.lower() != 'stable':
            if readiness not in ('preview',):
                raise exceptions.InvalidSwaggerValueError(
                    f"Invalid readiness value '{readiness}' in file path", file_path)
            file_path_version.readiness = ResourceVersion.Readiness.Preview
        return file_path_version

    def to_cmd(self, **kwargs):
        resource = CMDResource()
        resource.id = self.id
        resource.version = self.version

        resource.swagger = f"{self.resource_provider}/Paths/{b64encode_str(self.path)}/V/{b64encode_str(self.version)}"
        return resource


class ResourceVersion:
    class Readiness(enum.Enum):
        Preview = 'preview'
        Stable = 'stable'

    def __init__(self, version):
        readiness = self.Readiness.Stable
        for keyword in ('beta', 'preview', 'privatepreview'):
            if keyword in version.lower():
                readiness = self.Readiness.Preview

        self.version = version
        self.readiness = readiness

        self.date = datetime.date.min

        date_piece = version.split('.')[0]  # date is always the prefix of version
        if '-' in date_piece:
            pieces = date_piece.replace('_', '-').split('-')
            try:
                self.date = datetime.date.fromisoformat('-'.join(pieces[:3]))
            except ValueError as err:
                logger.warning(f'ParseVersionDateError: Version={version} : {err}')

    def __str__(self):
        return self.version

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return str(self) != str(other)



import logging
import abc

from schematics.models import Model
from schematics.types import ModelType, ListType, StringType, UTCDateTimeType, PolyModelType
from schematics.types.serializable import serializable
from utils.fields import PlaneField, CloudField
from urllib.parse import urlparse
from utils import exceptions


from ._arg_group import CMDArgGroup
from ._utils import CMDArgBuildPrefix, CMDDiffLevelEnum
from ._arg_builder import CMDArgBuilder
from ._schema import CMDSchemaField, CMDStringSchema
from ._subresource_selector import CMDSubresourceSelector
from ._operation import CMDHttpOperation
from ._command import handle_duplicated_options
from ._resource import CMDResource


logger = logging.getLogger('aaz')


class CMDClientAADAuthConfig(Model):
    scopes = ListType(StringType, required=True, min_size=1)

    class Options:
        serialize_when_none = False
    
    def reformat(self, **kwargs):
        # TODO: check scopes schema
        self.scopes = sorted(self.scopes)

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if sorted(self.scopes) != sorted(old.scopes):
                diff['scopes'] = f"{old.scopes} != {self.scopes}"
        return diff


class CMDClientAuth(Model):
    aad = ModelType(
        CMDClientAADAuthConfig,
    )

    class Options:
        serialize_when_none = False
    
    def reformat(self, **kwargs):
        if self.aad:
            self.aad.reformat(**kwargs)
        else:
            raise exceptions.VerificationError('Invalid auth config', default='Client auth is not defined')

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if old.aad:
                if not self.aad:
                    diff['aad'] = f"miss aad auth now."
                else:
                    aad_diff = self.aad.diff(old.aad, level)
                    if aad_diff:
                        diff['aad'] = aad_diff

        if level >= CMDDiffLevelEnum.Structure:
            if self.aad:
                aad_diff = self.aad.diff(old.aad, level)
                if aad_diff:
                    diff['aad'] = aad_diff
        return diff


class _EndpointTemplateMixin:
    @staticmethod
    def _iter_placeholders(template):
        endpoint = urlparse(template).netloc
        while True:
            idx = 0
            required = True
            while idx < len(endpoint) and endpoint[idx] != '{':
                idx += 1
            endpoint = endpoint[idx + 1:]
            if not endpoint:
                # not found '{'
                return

            idx = 0
            while idx < len(endpoint) and endpoint[idx] != '}':
                idx += 1
            if idx >= len(endpoint):
                # not found '}'
                return
            placeholder = endpoint[:idx]
            endpoint = endpoint[idx+1:]

            yield placeholder, required

    @staticmethod
    def _reformat(template):
        parsed = urlparse(template)
        if parsed.path:
            if parsed.path == '/' and not parsed.params and not parsed.query and not parsed.fragment:
                return template.rstrip('/')
            else:
                raise exceptions.VerificationError('Invalid endpoints', details='"{}" contains path'.format(template))
        if not parsed.scheme or not parsed.netloc:
            raise exceptions.VerificationError('Invalid endpoints', details='"{}" has no schema or hostname'.format(template))

        return template


class CMDClientEndpointTemplate(Model, _EndpointTemplateMixin):

    cloud = CloudField(required=True)
    
    # https://{accountName}.{zone}.blob.storage.azure.net
    template = StringType(required=True)

    class Options:
        serialize_when_none = False

    def iter_placeholders(self):
        return self._iter_placeholders(self.template)

    def reformat(self, **kwargs):
        self.template = self._reformat(self.template)

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.cloud != old.cloud:
                diff['cloud'] = f"{old.cloud} != {self.cloud}"
            if self.template != old.template:
                diff['template'] = f"{old.template} != {self.template}"
        return diff


class CMDClientEndpointCloudMetadataTemplate(Model, _EndpointTemplateMixin):

    selector_index = StringType(
        required=True,
        serialized_name="selectorIndex",
        deserialize_from="selectorIndex",
    )  # index, used to retrieve property in arm cloud metadata endpoints.
    prefix_template = StringType(
        serialized_name="prefixTemplate",
        deserialize_from="prefixTemplate",
    )  # prefix template, required for suffixes.

    class Options:
        serialize_when_none = False

    def iter_placeholders(self):
        if self.prefix_template is None:
            return
        for placeholder in self._iter_placeholders(self.prefix_template):
            yield placeholder

    def reformat(self, **kwargs):
        if self.prefix_template is None:
            return
        self.prefix_template = self._reformat(self.prefix_template)

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if self.selector_index != old.selector_index:
                diff['selector_index'] = f"{old.selector_index} != {self.selector_index}"
            if self.prefix_template != old.prefix_template:
                diff['prefix_template'] = f"{old.prefix_template} != {self.prefix_template}"
        return diff


class CMDClientEndpoints(Model):
    # properties as tags
    TYPE_VALUE = None  # types: "template", "http-operation"

    class Options:
        serialize_when_none = False

    @serializable
    def type(self):
        return self._get_type()

    def _get_type(self):
        assert self.TYPE_VALUE is not None
        return self.TYPE_VALUE

    @classmethod
    def _claim_polymorphic(cls, data):
        if cls.TYPE_VALUE is None:
            return False

        if isinstance(data, dict):
            type_value = data.get('type', None)
            return type_value == cls.TYPE_VALUE
        elif isinstance(data, CMDClientEndpoints):
            return data.TYPE_VALUE == cls.TYPE_VALUE
        return False

    @abc.abstractmethod
    def reformat(self, **kwargs):
        pass

    @abc.abstractmethod
    def prepare(self):
        pass

    @abc.abstractmethod
    def generate_args(self, ref_args):
        pass

    @abc.abstractmethod
    def diff(self, old, level):
        pass


class CMDClientEndpointsByTemplate(CMDClientEndpoints):
    TYPE_VALUE = 'template'

    templates = ListType(ModelType(CMDClientEndpointTemplate), required=True, min_size=1)
    cloud_metadata = ModelType(CMDClientEndpointCloudMetadataTemplate,
                               serialized_name="cloudMetadata",
                               deserialize_from="cloudMetadata")
    params = ListType(CMDSchemaField())

    def reformat(self, **kwargs):
        for template in self.templates:
            template.reformat(**kwargs)
        self.templates = sorted(self.templates, key=lambda e: e.cloud)
        if self.cloud_metadata:
            self.cloud_metadata.reformat(**kwargs)

        # make sure the placeholders across all the endpoints are consistent
        placeholders = {}
        for template in self.templates:
            for placeholder, required in template.iter_placeholders():
                if placeholder in placeholders:
                    placeholders[placeholder]['count'] += 1
                    if placeholders[placeholder]['required'] != required:
                        raise exceptions.VerificationError('Invalid endpoints', details='Inconsistent required for placeholder "{}"'.format(placeholder))
                else:
                    placeholders[placeholder] = {
                        "required": required,
                        "count": 1
                    }
        expected_count = len(self.templates)
        if self.cloud_metadata:
            for placeholder, required in self.cloud_metadata.iter_placeholders():
                if placeholder in placeholders:
                    placeholders[placeholder]['count'] += 1
                    if placeholders[placeholder]['required'] != required:
                        raise exceptions.VerificationError('Invalid endpoints', details='Inconsistent required for placeholder "{}" in endpoints or cloud metadata'.format(placeholder))
                else:
                    placeholders[placeholder] = {
                        "required": required,
                        "count": 1
                    }
            expected_count += 1
        for item in placeholders.values():
            if item['count'] != expected_count:
                raise exceptions.VerificationError('Invalid endpoints', details='placeholder "{}" is missed in some endpoints or cloud metadata'.format(placeholder))

        # make sure the parameters are consistent with the placeholders
        if self.params:
            for param in self.params:
                if param.name not in placeholders:
                    raise exceptions.VerificationError('Invalid endpoints', details='Unknown endpoint parameter: "{}"'.format(param.name))
                if param.required != placeholders[param.name]['required']:
                    raise exceptions.VerificationError('Invalid endpoints', details='Inconsistent required for parameter: "{}"'.format(param.name))
        if len(placeholders) > 0 and (not self.params or len(placeholders) != len(self.params)):
            raise exceptions.VerificationError('Invalid endpoints', details='Inconsistent endpoint templates and parameters')

        if self.params:
            self.params = sorted(self.params, key=lambda p: p.name)
    
    def prepare(self):
        params = {}
        for template in self.templates:
            for placeholder, required in template.iter_placeholders():
                if placeholder not in params:
                    params[placeholder] = CMDStringSchema({
                        "name": placeholder,
                        "required": required,
                        "skip_url_encoding": True,
                    })
        if self.cloud_metadata:
            for placeholder, required in self.cloud_metadata.iter_placeholders():
                if placeholder not in params:
                    params[placeholder] = CMDStringSchema({
                        "name": placeholder,
                        "required": required,
                        "skip_url_encoding": True,
                    })
        self.params = sorted(params.values(), key=lambda p: p.name) or None

    def generate_args(self, ref_args):
        args = []
        if self.params:
            for param in self.params:
                builder = CMDArgBuilder.new_builder(
                    schema=param,
                    var_prefix=CMDArgBuildPrefix.ClientEndpoint, ref_args=ref_args
                )
                args.extend(builder.get_args())
        return args

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if (not self.cloud_metadata) != (not old.cloud_metadata):
                diff['cloud_metadata'] = "Add cloud metadata" if self.cloud_metadata else "Remove cloud metadata"
            elif self.cloud_metadata:
                if cloud_metadata_diff := self.cloud_metadata.diff(old.cloud_metadata, level):
                    diff['cloud_metadata'] = cloud_metadata_diff

            if len(self.templates) != len(old.templates):
                diff['templates'] = "template not match"
            else:
                templates_diff = {}
                for template in self.templates:
                    peer_template = None
                    for old_template in old.templates:
                        if old_template.cloud == template.cloud:
                            peer_template = old_template
                            break
                    template_diff = template.diff(peer_template, level)
                    if template_diff:
                        templates_diff[template.cloud] = template_diff
                if templates_diff:
                    diff['templates'] = templates_diff
        return diff


class CMDClientEndpointsByHttpOperation(CMDClientEndpoints):
    TYPE_VALUE = 'http-operation'

    resource = ModelType(CMDResource, required=True)

    selector = PolyModelType(
        CMDSubresourceSelector,
        allow_subclasses=True,
        serialized_name="selector",
        deserialize_from="selector",
    )

    operation = ModelType(
        CMDHttpOperation,
        required=True,
        serialized_name="operation",
        deserialize_from="operation"
    )

    def reformat(self, **kwargs):
        self.selector.reformat(**kwargs)
        schema_cls_map = {}
        self.operation.reformat(schema_cls_map=schema_cls_map, **kwargs)
        for key, value in schema_cls_map.items():
            if value is None:
                raise exceptions.VerificationError(
                    message=f"ReformatError: Schema Class '{key}' not defined.",
                    details=None
                )

    def prepare(self):
        # TODO: remove unused schema based on selector
        pass

    def generate_args(self, ref_args):
        arguments = {}
        has_subresource = False
        if self.selector:
            has_subresource = True
            for arg in self.selector.generate_args(
                    ref_args=ref_args,
                    var_prefix=CMDArgBuildPrefix.ClientEndpoint
            ):
                if arg.var not in arguments:
                    arguments[arg.var] = arg

        for arg in self.operation.generate_args(
                ref_args=ref_args,
                has_subresource=has_subresource,
                var_prefix=CMDArgBuildPrefix.ClientEndpoint
        ):
            if arg.var not in arguments:
                arguments[arg.var] = arg

        return handle_duplicated_options(
            arguments, has_subresource=has_subresource, operation_id=self.operation.operation_id)

    def diff(self, old, level):
        if type(self) is not type(old):
            return f"Type: {type(old)} != {type(self)}"
        diff = {}
        if resource_diff := self.resource.diff(old.resource, level):
            diff["resource"] = resource_diff
        if selector_diff := self.selector.diff(old.selector, level):
            diff["selector"] = selector_diff
        if operation_diff := self.operation.diff(old.operation, level):
            diff["operation"] = operation_diff
        return diff


class CMDClientConfig(Model):
    # this property is used to manage the client config version.
    version = UTCDateTimeType(required=True)

    plane = PlaneField(required=True)
    endpoints = PolyModelType(
        CMDClientEndpoints,
        allow_subclasses=True,
        required=True)

    auth = ModelType(CMDClientAuth, required=True)
    arg_group = ModelType(
        CMDArgGroup,
        serialized_name='argGroup',
        deserialize_from='argGroup',
    )

    class Options:
        serialize_when_none = False

    def reformat(self, **kwargs):
        self.endpoints.reformat(**kwargs)
        self.auth.reformat(**kwargs)
        if self.arg_group:
            self.arg_group.reformat(**kwargs)

    def generate_args(self, ref_args=None, ref_options=None):
        if not ref_args:
            ref_args = []
            if self.arg_group:
                ref_args.extend(self.arg_group.args)
            ref_args = ref_args or None

        arguments = {}
        for arg in self.endpoints.generate_args(ref_args):
            if arg.var not in arguments:
                if ref_options and arg.var in ref_options:
                    # replace generated options by ref_options
                    arg.options = [*ref_options[arg.var]]
                arg.group = 'Client'
                arguments[arg.var] = arg

        # verify duplicated options
        used_args = set()
        for arg in arguments.values():
            used_args.add(arg.var)
            r_arg = None
            for v in arguments.values():
                if v.var in used_args:
                    continue
                if not set(arg.options).isdisjoint(v.options):
                    r_arg = v
                    break
            if r_arg:
                raise exceptions.VerificationError(
                    "Argument Options conflict",
                    details=f"Duplicated Option Value: {set(arg.options).intersection(r_arg.options)} : {arg.var} with {r_arg.var}"
                )

        if arguments:
            self.arg_group = CMDArgGroup({
                "name": 'Client',
            })
            self.arg_group.args = sorted(arguments.values(), key=lambda a: a.var)
        else:
            self.arg_group = None



import os
import logging
import re

logger = logging.getLogger('aaz')


class TypeSpecHelper:

    @staticmethod
    def _iter_entry_files(folder):
        if not os.path.isdir(folder):
            raise ValueError(f"Path not exist: {folder}")
        ts_path = os.path.join(folder, "main.tsp")
        cfg_path = os.path.join(folder, "tspconfig.yaml")
        if os.path.isfile(ts_path) and os.path.isfile(cfg_path):
            yield ts_path, cfg_path
        for root, dirs, files in os.walk(folder):
            ts_path = os.path.join(root, "main.tsp")
            cfg_path = os.path.join(root, "tspconfig.yaml")
            if os.path.isfile(ts_path) and os.path.isfile(cfg_path):
                yield ts_path, cfg_path

    @classmethod
    def find_mgmt_plane_entry_files(cls, folder):
        files = []
        for ts_path, cfg_path in cls._iter_entry_files(folder):
            namespace, is_mgmt_plane = cls._parse_main_tsp(ts_path)
            if is_mgmt_plane:
                files.append((namespace, ts_path, cfg_path))
        return files

    @classmethod
    def find_data_plane_entry_files(cls, folder):
        files = []
        if not os.path.isdir(folder):
            return files
        for ts_path, cfg_path in cls._iter_entry_files(folder):
            namespace, is_mgmt_plane = cls._parse_main_tsp(ts_path)
            if not namespace:
                continue
            if not is_mgmt_plane:
                files.append((namespace, ts_path, cfg_path))
        return files

    @classmethod
    def _parse_main_tsp(cls, path):
        def load_file_lines(path):
            with open(path, "r", encoding="utf-8") as f:
                return f.readlines()

        def parse_by_line(lines):
            namespace, is_mgmt_plane = None, False
            for line in lines:
                if line.startswith("@armProviderNamespace"):
                    is_mgmt_plane = True

                if line.startswith("namespace "):
                    assert namespace is None
                    namespace = re.match(r"^namespace\s+([A-Za-z0-9.]+)", line).group(1)
                    # armProviderNamespace will always be appeared before namespace
                    break

            return namespace, is_mgmt_plane

        def expand(path):
            base = os.path.dirname(path)

            content = []
            for line in lines:
                if match := re.compile(r'^import "([^"]+)"').findall(line):
                    rel_path = match[0]
                    abs_path = os.path.abspath(os.path.join(base, rel_path))

                    if os.path.isfile(abs_path):  # expand first level only; otherwise, may have circular reference
                        with open(abs_path, "r", encoding="utf-8") as f:
                            content.append("".join(f.readlines()))
                    else:
                        content.append(line)
                else:
                    content.append(line)

            return "".join(content).split("\n")

        lines = load_file_lines(path)
        if any("@armProviderNamespace" in line for line in lines):
            namespace, is_mgmt_plane = parse_by_line(lines)

        else:
            namespace, is_mgmt_plane = parse_by_line(expand(path))

        if namespace is None:
            return None, None

        return namespace, is_mgmt_plane



from schematics.models import Model
from schematics.types import StringType, ModelType, BooleanType
from ..fields import BooleanType
from ..x_ms_pageable import XmsPageableField

from command.model.configuration import CMDHttpOperation


class TypeSpecOperation(Model):
    operation_id = StringType(
        serialized_name="operationId",
        deserialize_from="operationId",
    )  # Unique string used to identify the operation. The id MUST be unique among all operations described in the API. Tools and libraries MAY use the operationId to uniquely identify an operation, therefore, it is recommended to follow common programming naming conventions.

    x_ms_pageable = XmsPageableField(serialized_name="pageable", deserialize_from="pageable")
    # x_ms_examples = XmsExamplesField()  # TODO:

    # MutabilityEnum.Read
    read = ModelType(CMDHttpOperation)
    # MutabilityEnum.Create
    create = ModelType(CMDHttpOperation)
    # MutabilityEnum.Update
    update = ModelType(CMDHttpOperation)



from command.model.configuration import CMDStringArgBase, CMDByteArgBase, CMDBinaryArgBase, CMDDurationArgBase, \
    CMDDateArgBase, CMDDateTimeArgBase, CMDTimeArgBase, CMDAnyTypeArgBase, CMDUuidArgBase, CMDPasswordArgBase, \
    CMDSubscriptionIdArgBase, CMDResourceGroupNameArgBase, CMDResourceIdArgBase, CMDResourceLocationArgBase, \
    CMDIntegerArgBase, CMDBooleanArgBase, CMDFloatArgBase, CMDObjectArgBase, CMDArrayArgBase, CMDClsArgBase, \
    CMDSubscriptionIdArg, CMDArg
from command.model.configuration import CMDArgGroup, CMDArgumentHelp
from command.model.configuration import CMDStringFormat, CMDIntegerFormat, CMDFloatFormat, CMDObjectFormat, \
    CMDArrayFormat, CMDDateTimeFormat
from utils.case import to_camel_case, to_snake_case
from utils import exceptions
from utils.stage import AAZStageEnum


class AzArgGroupGenerator:

    def __init__(self, args_schema_name, cmd_ctx, arg_group):
        assert isinstance(arg_group, CMDArgGroup)
        assert arg_group.name is not None   # empty string is valid
        self.name = arg_group.name
        self._arg_group = arg_group
        self._args_schema_name = args_schema_name
        self._cmd_ctx = cmd_ctx

        # update args
        for arg in self._arg_group.args:
            self._update_over_arg(arg, parse_arg_name(arg))

    def _update_over_arg(self, arg, *arg_keys):
        if isinstance(arg, CMDArg):
            if isinstance(arg, CMDSubscriptionIdArg) \
                    and arg_keys == ('subscription', ) and 'subscription' in arg.options:
                # use self.ctx.subscription_id
                self._cmd_ctx.set_argument(('subscription_id', ), arg, ctx_namespace='self.ctx')
            else:
                self._cmd_ctx.set_argument(arg_keys, arg)

        if getattr(arg, 'cls', None):
            self._cmd_ctx.set_argument_cls(arg)
            arg_keys = (f"@{arg.cls}", )  # prepare for cls sub property use

        if isinstance(arg, CMDObjectArgBase):
            if arg.args:
                for sub_arg in arg.args:
                    self._update_over_arg(sub_arg, *arg_keys, parse_arg_name(sub_arg))
            if arg.additional_props and arg.additional_props.item:
                self._update_over_arg(arg.additional_props.item, *arg_keys, '{}')
        elif isinstance(arg, CMDArrayArgBase):
            if arg.item:
                self._update_over_arg(arg.item, *arg_keys, '[]')

    def iter_scopes(self):
        scope = self._args_schema_name
        scope_define = f"cls.{self._args_schema_name}"
        rendered_args = []
        search_args = {}
        for a in self._arg_group.args:
            if a.hide:
                # escape hide argument
                continue

            if isinstance(a, CMDSubscriptionIdArg) and 'subscription' in a.options:
                # ignore subscription id, because cli core registered global _subscription argument
                continue

            a_name = parse_arg_name(a)
            a_type, a_kwargs, cls_builder_name = render_arg(a, self._cmd_ctx, arg_group=self.name)

            rendered_args.append((a_name, a_type, a_kwargs, cls_builder_name))
            if not cls_builder_name and isinstance(a, (CMDObjectArgBase, CMDArrayArgBase)):
                search_args[a_name] = a
        if rendered_args:
            yield scope, scope_define, rendered_args

        for a_name, a in search_args.items():
            for scopes in _iter_scopes_by_arg_base(a, a_name, f"{scope_define}.{a_name}", self._cmd_ctx):
                yield scopes


class AzArgClsGenerator:

    def __init__(self, name, cmd_ctx, arg):
        self.arg = arg
        self.name = name
        self.args_schema_name = f"_args_{to_snake_case(name)}"
        self.builder_name = parse_cls_builder_name(name)
        self._cmd_ctx = cmd_ctx
        self.arg_type, self.arg_kwargs, _ = render_arg_base(self.arg, self._cmd_ctx)

        self.props = []
        if isinstance(arg, CMDObjectArgBase):
            if arg.args and arg.additional_props:
                # treated as AAZFreeFormDictArg
                pass
            elif arg.args:
                for a in arg.args:
                    if a.hide:
                        # escape hide argument
                        continue
                    self.props.append(parse_arg_name(a))
            elif arg.additional_props:
                if arg.additional_props.item is not None:
                    self.props.append("Element")
        elif isinstance(arg, CMDArrayArgBase):
            self.props.append("Element")

        self.props = sorted(self.props)

    def iter_scopes(self):
        for scopes in _iter_scopes_by_arg_base(self.arg, to_snake_case(self.name), f"cls.{self.args_schema_name}", self._cmd_ctx):
            yield scopes


def _iter_scopes_by_arg_base(arg, name, scope_define, cmd_ctx):
    rendered_args = []
    search_args = {}

    if isinstance(arg, CMDObjectArgBase):
        if arg.args and arg.additional_props:
            # treated as AAZFreeFormDictArg so no need to iterate
            pass
        elif arg.args:
            for a in arg.args:
                if a.hide:
                    # escape hide argument
                    continue
                a_name = parse_arg_name(a)
                a_type, a_kwargs, cls_builder_name = render_arg(a, cmd_ctx)
                rendered_args.append((a_name, a_type, a_kwargs, cls_builder_name))
                if not cls_builder_name and isinstance(a, (CMDObjectArgBase, CMDArrayArgBase)):
                    search_args[a_name] = a
        elif arg.additional_props:
            if arg.additional_props.item is not None:
                # AAZDictArg
                a = arg.additional_props.item
                a_name = "Element"
                a_type, a_kwargs, cls_builder_name = render_arg_base(a, cmd_ctx)
                rendered_args.append((a_name, a_type, a_kwargs, cls_builder_name))
                if not cls_builder_name and isinstance(a, (CMDObjectArgBase, CMDArrayArgBase)):
                    search_args[a_name] = a
            else:
                # AAZFreeFormDictArg
                # TODO: remove it after aaz repo updated without any_type property
                assert arg.additional_props.any_type is True
    elif isinstance(arg, CMDArrayArgBase):
        # AAZListArg
        assert arg.item is not None
        a = arg.item
        a_name = "Element"
        a_type, a_kwargs, cls_builder_name = render_arg_base(a, cmd_ctx)
        rendered_args.append((a_name, a_type, a_kwargs, cls_builder_name))
        if not cls_builder_name and isinstance(a, (CMDObjectArgBase, CMDArrayArgBase)):
            search_args[a_name] = a
    else:
        raise NotImplementedError()

    if rendered_args:
        yield name, scope_define, rendered_args

    for a_name, a in search_args.items():
        a_scope_define = f"{scope_define}.{a_name}"
        if a_name == "Element":
            a_name = '_element'
        for scopes in _iter_scopes_by_arg_base(a, a_name, a_scope_define, cmd_ctx):
            yield scopes


def parse_cls_builder_name(cls_name):
    return f"_build_args_{to_snake_case(cls_name)}"


def parse_arg_help(help):
    assert isinstance(help, CMDArgumentHelp)
    if not help.lines and not help.ref_commands:
        if not help.short:
            raise exceptions.InvalidAPIUsage("Invalid argument help, short summary is miss.")
        return help.short
    h = {
        "short-summary": help.short
    }
    if help.lines:
        h["long-summary"] = '\n'.join(help.lines)
    if help.ref_commands:
        h["populator-commands"] = [*help.ref_commands]
    return h


def parse_arg_enum(enum):
    if not enum or not enum.items:
        return None
    e = {}
    for item in enum.items:
        if item.hide:
            continue
        e[item.name] = item.value
    kwargs = {
        "enum": e
    }
    if enum.support_extension:
        kwargs['enum_support_extension'] = True
    return kwargs


def parse_arg_name(arg):
    arg_name_length = 0
    arg_name = None
    for option in arg.options:
        if len(option) > arg_name_length:
            arg_name = to_snake_case(option)
            arg_name_length = len(option)
    return arg_name


def render_arg(arg, cmd_ctx, arg_group=None):
    arg_kwargs = {
        "options": []
    }
    for option in arg.options:
        if arg_group is not None:
            if len(option) == 1:
                option = f"-{option}"
            else:
                option = f"--{option}"
        arg_kwargs["options"].append(option)
    if getattr(arg, 'singular_options', None):
        arg_kwargs["singular_options"] = []
        for option in arg.singular_options:
            if arg_group is not None:
                if len(option) == 1:
                    option = f"-{option}"
                else:
                    option = f"--{option}"
            arg_kwargs["singular_options"].append(option)

    if arg_group:  # not empty string or None
        arg_kwargs["arg_group"] = arg_group
    if arg.help:
        arg_kwargs["help"] = parse_arg_help(arg.help)
    if arg.required:
        arg_kwargs["required"] = True
    if arg.stage == AAZStageEnum.Preview:
        arg_kwargs["is_preview"] = True
    elif arg.stage == AAZStageEnum.Experimental:
        arg_kwargs["is_experimental"] = True

    if arg.id_part and cmd_ctx.support_id_part:
        arg_kwargs["id_part"] = arg.id_part

    if arg.default:
        arg_kwargs["default"] = arg.default.value

    if arg.configuration_key:
        arg_kwargs["configured_default"] = arg.configuration_key

    arg_type, arg_kwargs, cls_builder_name = render_arg_base(arg, cmd_ctx, arg_kwargs)

    if arg.prompt:
        if arg_type == "AAZPasswordArg":
            arg_kwargs["prompt"] = {
                "cls": "AAZPromptPasswordInput",
                "kwargs": {
                    "msg": arg.prompt.msg,
                }
            }
            if arg.prompt.confirm:
                arg_kwargs["prompt"]["kwargs"]["confirm"] = arg.prompt.confirm
        else:
            arg_kwargs["prompt"] = {
                "cls": "AAZPromptInput",
                "kwargs": {
                    "msg": arg.prompt.msg,
                }
            }
    if "blank" in arg_kwargs and "prompt" in arg_kwargs:
        raise KeyError("An argument cannot contain both prompt and blank")
    return arg_type, arg_kwargs, cls_builder_name


def render_arg_base(arg, cmd_ctx, arg_kwargs=None):
    if isinstance(arg, CMDClsArgBase):
        cls_name = arg.type[1:]
        arg = cmd_ctx.arg_clses[cls_name].arg
    else:
        cls_name = getattr(arg, 'cls', None)
    cls_builder_name = parse_cls_builder_name(cls_name) if cls_name else None

    if arg_kwargs is None:
        arg_kwargs = {}

    if arg.nullable:
        arg_kwargs['nullable'] = True

    if arg.blank:
        arg_kwargs["blank"] = arg.blank.value

    if isinstance(arg, CMDAnyTypeArgBase):
        arg_type = "AAZAnyTypeArg"
    elif isinstance(arg, CMDStringArgBase):
        arg_type = "AAZStrArg"
        enum_kwargs = parse_arg_enum(arg.enum)
        if enum_kwargs:
            arg_kwargs.update(enum_kwargs)

        if arg.fmt and isinstance(arg.fmt, CMDStringFormat):
            arg_kwargs['fmt'] = fmt = {
                "cls": "AAZStrArgFormat",
                "kwargs": {}
            }
            if arg.fmt.pattern is not None:
                fmt['kwargs']["pattern"] = arg.fmt.pattern
            if arg.fmt.max_length is not None:
                fmt['kwargs']["max_length"] = arg.fmt.max_length
            if arg.fmt.min_length is not None:
                fmt['kwargs']["min_length"] = arg.fmt.min_length

        if isinstance(arg, CMDSubscriptionIdArgBase):
            arg_type = "AAZSubscriptionIdArg"
        elif isinstance(arg, CMDResourceGroupNameArgBase):
            arg_type = "AAZResourceGroupNameArg"
            if 'options' in arg_kwargs and set(arg_kwargs['options']) == {'--resource-group', '-g'}:
                # it's default value
                del arg_kwargs['options']
            if 'id_part' in arg_kwargs and arg_kwargs['id_part'] == 'resource_group':
                # it's default value
                del arg_kwargs['id_part']
        elif isinstance(arg, CMDResourceIdArgBase):
            arg_type = "AAZResourceIdArg"
            if arg.fmt:
                arg_kwargs['fmt'] = fmt = {
                    "cls": "AAZResourceIdArgFormat",
                    "kwargs": {}
                }
                if arg.fmt.template is not None:
                    fmt['kwargs']['template'] = cmd_ctx.render_arg_resource_id_template(arg.fmt.template)

        elif isinstance(arg, CMDResourceLocationArgBase):
            arg_type = "AAZResourceLocationArg"
            if 'options' in arg_kwargs and set(arg_kwargs['options']) == {'--location', '-l'}:
                # it's default value
                del arg_kwargs['options']
            if not arg.nullable and not arg.no_rg_default and cmd_ctx.rg_arg_var:
                resource_group_arg, hide = cmd_ctx.get_argument(cmd_ctx.rg_arg_var)
                if not hide:
                    resource_group_arg = resource_group_arg.replace('self.ctx.args.', '')
                    arg_kwargs['fmt'] = fmt = {
                        "cls": "AAZResourceLocationArgFormat",
                        "kwargs": {
                            "resource_group_arg": resource_group_arg
                        }
                    }
        elif isinstance(arg, CMDByteArgBase):
            raise NotImplementedError()
        elif isinstance(arg, CMDBinaryArgBase):
            raise NotImplementedError()
        elif isinstance(arg, CMDDurationArgBase):
            arg_type = "AAZDurationArg"
        elif isinstance(arg, CMDDateArgBase):
            arg_type = "AAZDateArg"
        elif isinstance(arg, CMDDateTimeArgBase):
            arg_type = "AAZDateTimeArg"
            if arg.fmt and isinstance(arg.fmt, CMDDateTimeFormat):
                arg_kwargs['fmt'] = {
                    "cls": "AAZDateTimeFormat",
                    "kwargs": {
                        "protocol": arg.fmt.protocol
                    }
                }
        elif isinstance(arg, CMDTimeArgBase):
            arg_type = "AAZTimeArg"
        elif isinstance(arg, CMDUuidArgBase):
            arg_type = "AAZUuidArg"
        elif isinstance(arg, CMDPasswordArgBase):
            arg_type = "AAZPasswordArg"

    elif isinstance(arg, CMDIntegerArgBase):
        arg_type = "AAZIntArg"
        enum_kwargs = parse_arg_enum(arg.enum)
        if enum_kwargs:
            arg_kwargs.update(enum_kwargs)

        if arg.fmt and isinstance(arg.fmt, CMDIntegerFormat):
            arg_kwargs['fmt'] = fmt = {
                "cls": "AAZIntArgFormat",
                "kwargs": {}
            }
            if arg.fmt.multiple_of is not None:
                fmt['kwargs']["multiple_of"] = arg.fmt.multiple_of
            if arg.fmt.maximum is not None:
                fmt['kwargs']["maximum"] = arg.fmt.maximum
            if arg.fmt.minimum is not None:
                fmt['kwargs']["minimum"] = arg.fmt.minimum

        # TODO: add format for integer32 and integer64
        # if isinstance(arg, CMDInteger32ArgBase):
        #     raise NotImplementedError()
        # elif isinstance(arg, CMDInteger64ArgBase):
        #     raise NotImplementedError()

    elif isinstance(arg, CMDBooleanArgBase):
        arg_type = "AAZBoolArg"
        if arg.reverse:
            arg_kwargs['fmt'] = {
                "cls": "AAZBoolArgFormat",
                "kwargs": {
                    "reverse": arg.reverse,
                }
            }

    elif isinstance(arg, CMDFloatArgBase):
        arg_type = "AAZFloatArg"
        enum_kwargs = parse_arg_enum(arg.enum)
        if enum_kwargs:
            arg_kwargs.update(enum_kwargs)

        if arg.fmt and isinstance(arg.fmt, CMDFloatFormat):
            arg_kwargs['fmt'] = fmt = {
                "cls": "AAZFloatArgFormat",
                "kwargs": {}
            }
            if arg.fmt.multiple_of is not None:
                fmt['kwargs']["multiple_of"] = arg.fmt.multiple_of
            if arg.fmt.maximum is not None:
                fmt['kwargs']['maximum'] = arg.fmt.maximum
            if arg.fmt.minimum is not None:
                fmt['kwargs']['minimum'] = arg.fmt.minimum
            if arg.fmt.exclusive_maximum is not None:
                fmt['kwargs']['exclusive_maximum'] = arg.fmt.exclusive_maximum
            if arg.fmt.exclusive_minimum is not None:
                fmt['kwargs']['exclusive_minimum'] = arg.fmt.exclusive_minimum

        # TODO: add format for float32 and float64
        # if isinstance(arg, CMDFloat32ArgBase):
        #     raise NotImplementedError()
        # elif isinstance(arg, CMDFloat64ArgBase):
        #     raise NotImplementedError()

    elif isinstance(arg, CMDObjectArgBase):
        if arg.additional_props:
            if arg.additional_props.any_type is True:
                arg_type = "AAZFreeFormDictArg"
                arg_fmt_cls = "AAZFreeFormDictArgFormat"
            elif arg.args and arg.additional_props:
                # not support object with both args and additional_props, so we treat it as AAZFreeFormDictArg
                arg_type = "AAZFreeFormDictArg"
                arg_fmt_cls = "AAZFreeFormDictArgFormat"
            else:
                assert arg.additional_props.item is not None
                arg_type = "AAZDictArg"
                arg_fmt_cls = "AAZDictArgFormat"
            if arg.fmt is not None:
                assert isinstance(arg.fmt, CMDObjectFormat)
                arg_kwargs['fmt'] = fmt = {
                    "cls": arg_fmt_cls,
                    "kwargs": {}
                }
                if arg.fmt.max_properties is not None:
                    fmt['kwargs']['max_properties'] = arg.fmt.max_properties
                if arg.fmt.min_properties is not None:
                    fmt['kwargs']['min_properties'] = arg.fmt.min_properties
        else:
            arg_type = "AAZObjectArg"
            if arg.additional_props:
                raise NotImplementedError()
            if arg.fmt is not None:
                assert isinstance(arg.fmt, CMDObjectFormat)
                arg_kwargs['fmt'] = fmt = {
                    "cls": "AAZObjectArgFormat",
                    "kwargs": {}
                }
                if arg.fmt.max_properties is not None:
                    fmt['kwargs']['max_properties'] = arg.fmt.max_properties
                if arg.fmt.min_properties is not None:
                    fmt['kwargs']['min_properties'] = arg.fmt.min_properties

    elif isinstance(arg, CMDArrayArgBase):
        arg_type = "AAZListArg"
        if arg.fmt is not None:
            arg_kwargs['fmt'] = fmt = {
                "cls": "AAZListArgFormat",
                "kwargs": {}
            }
            assert isinstance(arg.fmt, CMDArrayFormat)
            if arg.fmt.unique is not None:
                fmt['kwargs']['unique'] = arg.fmt.unique
            if arg.fmt.max_length is not None:
                fmt['kwargs']['max_length'] = arg.fmt.max_length
            if arg.fmt.min_length is not None:
                fmt['kwargs']['min_length'] = arg.fmt.min_length
    else:
        raise NotImplementedError()

    return arg_type, arg_kwargs, cls_builder_name



import os
import glob

from utils.plane import PlaneEnum
from swagger.utils.tools import resolve_path_to_uri
from ._resource_provider import OpenAPIResourceProvider, TypeSpecResourceProvider
from ._typespec_helper import TypeSpecHelper


class SwaggerModule:

    def __init__(self, plane, name, folder_path, parent=None):
        assert plane == PlaneEnum.Mgmt or PlaneEnum.is_data_plane(plane), f"Invalid plane: '{plane}'"
        self.plane = plane
        self.name = name
        self.folder_path = folder_path
        self._parent = parent

    def __str__(self):
        if self._parent is not None:
            return f"{self._parent}/{self.name}"
        else:
            return f"{self.name}"

    def __eq__(self, other):
        return isinstance(other, self.__class__) and str(other) == str(self)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(str(self))

    @property
    def names(self):
        if self._parent is None:
            return [self.name]
        else:
            return [*self._parent.names, self.name]


class MgmtPlaneModule(SwaggerModule):

    def __str__(self):
        if self._parent is None:
            return f'{self.plane}/{super().__str__()}'
        else:
            return super(MgmtPlaneModule, self).__str__()

    def get_resource_providers(self):
        rp = []
        rp.extend(self._get_openapi_resource_providers())
        rp.extend(self._get_typespec_resource_providers())
        return rp

    def _get_openapi_resource_providers(self):
        rp = []
        if 'resource-manager' not in self.folder_path:
            folder_path = os.path.join(self.folder_path, 'resource-manager')
        else:
            # ignore to add resource-manager folder for sub openapi modules
            folder_path = self.folder_path
        if not os.path.exists(folder_path):
            return rp
        for name in os.listdir(folder_path):
            path = os.path.join(folder_path, name)
            if os.path.isdir(path):
                name_parts = name.split('.')
                if len(name_parts) >= 2:
                    readme_paths = [*_search_readme_md_paths(path, search_parent=True)]
                    rp.append(OpenAPIResourceProvider(name, path, readme_paths, swagger_module=self))
                elif name.lower() != 'common':
                    # azsadmin module only
                    sub_module = MgmtPlaneModule(plane=self.plane, name=name, folder_path=path, parent=self)
                    rp.extend(sub_module.get_resource_providers())
        return rp

    def _get_typespec_resource_providers(self):
        rp = {}
        if 'resource-manager' in self.folder_path:
            return rp

        for namespace, ts_path, cfg_path in TypeSpecHelper.find_mgmt_plane_entry_files(self.folder_path):
            entry_file = resolve_path_to_uri(ts_path)
            if namespace in rp:
                rp[namespace].entry_files.append(entry_file)
            else:
                rp[namespace] = TypeSpecResourceProvider(
                    name=namespace, entry_files=[entry_file], swagger_module=self)
        return [*rp.values()]


class DataPlaneModule(SwaggerModule):

    def __str__(self):
        if self._parent is None:
            return f'{self.plane}/{super().__str__()}'
        else:
            return super(DataPlaneModule, self).__str__()

    def get_resource_providers(self):
        rp = []
        rp.extend(self._get_openapi_resource_providers())
        rp.extend(self._get_typespec_resource_providers())
        scope = PlaneEnum.get_data_plane_scope(self.plane)
        if scope:
            rp = [r for r in rp if r.name.lower() == scope]
        return rp

    def _get_openapi_resource_providers(self):
        rp = []
        if 'data-plane' not in self.folder_path:
            folder_path = os.path.join(self.folder_path, 'data-plane')
        else:
            # ignore to add data-plane folder for sub openapi modules
            folder_path = self.folder_path
        if not os.path.exists(folder_path):
            return rp
        for name in os.listdir(folder_path):
            path = os.path.join(folder_path, name)
            if os.path.isdir(path):
                if name.lower() in ('preview', 'stable'):
                    continue
                name_parts = name.split('.')
                if len(name_parts) >= 2:
                    readme_paths = [*_search_readme_md_paths(path, search_parent=True)]
                    rp.append(OpenAPIResourceProvider(name, path, readme_paths, swagger_module=self))
                elif name.lower() != 'common':
                    sub_module = DataPlaneModule(plane=self.plane, name=name, folder_path=path, parent=self)
                    rp.extend(sub_module.get_resource_providers())
        return rp

    def _get_typespec_resource_providers(self):
        rp = {}
        if 'resource-manager' in self.folder_path:
            return rp

        for namespace, ts_path, cfg_path in TypeSpecHelper.find_data_plane_entry_files(self.folder_path):
            entry_file = resolve_path_to_uri(ts_path)
            if namespace in rp:
                rp[namespace].entry_files.append(entry_file)
            else:
                rp[namespace] = TypeSpecResourceProvider(
                    name=namespace, entry_files=[entry_file], swagger_module=self)
        return [*rp.values()]


def _search_readme_md_paths(path, search_parent=False):
    # Check parent directory first if requested
    if search_parent:
        parent_readme = os.path.join(os.path.dirname(path), 'readme.md')
        if os.path.isfile(parent_readme):
            yield parent_readme

    # Use glob to recursively find all readme.md files
    pattern = os.path.join(path, '**', 'readme.md')
    yield from glob.iglob(pattern, recursive=True)



from swagger.controller._example_builder import SwaggerExampleBuilder
from swagger.model.schema.example_item import XmsExamplesField
from swagger.model.schema.path_item import PathItem
from swagger.model.specs import SwaggerLoader


class ExampleGenerator:
    def __init__(self):
        self.loader = SwaggerLoader()

    def load_examples(self, resources):
        for resource in resources:
            self.loader.load_file(resource.file_path)
            self.loader.link_swaggers()

    def create_draft_examples_by_swagger(self, resources, command, cmd_operation_ids, cmd_name):
        cmd_examples = []

        for resource in resources:
            swagger = self.loader.get_loaded(resource.file_path)
            if not swagger:
                continue

            path_item = swagger.paths.get(resource.path, None)
            if path_item is None:
                path_item = swagger.x_ms_paths.get(resource.path, None)
            if not isinstance(path_item, PathItem):
                continue

            example_builder = None
            examples = None
            if path_item.get is not None and path_item.get.operation_id in cmd_operation_ids:
                example_builder = SwaggerExampleBuilder(
                    command=command,
                    path_item=path_item,
                    operation=path_item.get,
                    cmd_operation=cmd_operation_ids[path_item.get.operation_id]
                )
                examples = path_item.get.x_ms_examples

            elif path_item.delete is not None and path_item.delete.operation_id in cmd_operation_ids:
                example_builder = SwaggerExampleBuilder(
                    command=command,
                    path_item=path_item,
                    operation=path_item.delete,
                    cmd_operation=cmd_operation_ids[path_item.delete.operation_id]
                )
                examples = path_item.delete.x_ms_examples

            elif path_item.put is not None and path_item.put.operation_id in cmd_operation_ids:
                example_builder = SwaggerExampleBuilder(
                    command=command,
                    path_item=path_item,
                    operation=path_item.put,
                    cmd_operation=cmd_operation_ids[path_item.put.operation_id]
                )
                examples = path_item.put.x_ms_examples

            elif path_item.patch is not None and path_item.patch.operation_id in cmd_operation_ids:
                example_builder = SwaggerExampleBuilder(
                    command=command,
                    path_item=path_item,
                    operation=path_item.patch,
                    cmd_operation=cmd_operation_ids[path_item.patch.operation_id]
                )
                examples = path_item.patch.x_ms_examples

            elif path_item.post is not None and path_item.post.operation_id in cmd_operation_ids:
                example_builder = SwaggerExampleBuilder(
                    command=command,
                    path_item=path_item,
                    operation=path_item.post,
                    cmd_operation=cmd_operation_ids[path_item.post.operation_id]
                )
                examples = path_item.post.x_ms_examples

            elif path_item.head is not None and path_item.head.operation_id in cmd_operation_ids:
                example_builder = SwaggerExampleBuilder(
                    command=command,
                    path_item=path_item,
                    operation=path_item.head,
                    cmd_operation=cmd_operation_ids[path_item.head.operation_id]
                )
                examples = path_item.head.x_ms_examples

            if not example_builder or not examples:
                continue

            cmd_examples.extend(self.generate_examples(cmd_name, examples, example_builder))

        return cmd_examples

    @staticmethod
    def generate_examples(cmd_name, examples, example_builder):
        cmd_examples = []
        for name, example_item in examples.items():
            cmd_example = example_item.to_cmd(example_builder, cmd_name)
            if not cmd_example:
                continue

            cmd_example.name = name
            cmd_examples.append(cmd_example)

        return cmd_examples



from setuptools import find_packages, setup
from version import VERSION


def read_requirements(filename: str):
    with open(filename) as requirements_file:
        import re

        def fix_url_dependencies(req: str) -> str:
            """Pip and setuptools disagree about how URL dependencies should be handled."""
            m = re.match(
                r"^(git\+)?(https|ssh)://(git@)?github\.com/([\w-]+)/(?P<name>[\w-]+)\.git", req
            )
            if m is None:
                return req
            else:
                return f"{m.group('name')} @ {req}"

        requirements = []
        for line in requirements_file:
            line = line.strip()
            if line.startswith("#") or len(line) <= 0:
                continue
            requirements.append(fix_url_dependencies(line))
    return requirements


# # version.py defines the VERSION and VERSION_SHORT variables.
# # We use exec here, so we don't import cached_path whilst setting up.
# VERSION = {}  # type: ignore
# with open("version.py", "r") as version_file:
#     exec(version_file.read(), VERSION)

with open("README.md", "r", encoding="utf-8") as fp:
    README = fp.read()
with open("HISTORY.rst", "r", encoding="utf-8") as fp:
    HISTORY = fp.read()


setup(
    name='aaz-dev',
    version=VERSION,
    description='Microsoft Atomic Azure CLI Commands Developer Tools',
    long_description=README,
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14"
    ],
    keywords="azure",
    url="https://github.com/Azure/aaz-dev-tools",
    author="Microsoft Corporation",
    author_email="azpycli@microsoft.com",
    license='MIT',
    package_dir={
        "": "src"
    },
    packages=find_packages(
        where="src",
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"],
    ),
    include_package_data=True,
    install_requires=read_requirements("requirements.txt"),
    python_requires=">=3.10",
    entry_points={
        "console_scripts": ["aaz-dev=aaz_dev.app.main:main"]
    },
)



import logging
from urllib.parse import urljoin

from schematics.models import Model
from schematics.types import StringType, ModelType, ListType, DictType, BooleanType, PolyModelType

from command.model.configuration import CMDHttpOperation, CMDHttpAction, CMDHttpRequest, CMDHttpRequestPath, \
    CMDHttpRequestQuery, CMDHttpRequestHeader, CMDHttpRequestJsonBody, CMDRequestJson, CMDHttpOperationLongRunning
from swagger.utils import exceptions
from swagger.utils.tools import swagger_resource_path_to_resource_id_template
from .example_item import XmsExamplesField
from .external_documentation import ExternalDocumentation
from .fields import MimeField, XmsRequestIdField, SecurityRequirementField, XPublishField, \
    XSfCodeGenField, XmsClientNameField
from .parameter import ParameterField, PathParameter, QueryParameter, HeaderParameter, BodyParameter,\
    FormDataParameter, ParameterBase
from .reference import Reference, Linkable
from .response import Response
from .schema import ReferenceSchema
from .x_ms_long_running_operation import XmsLongRunningOperationField, XmsLongRunningOperationOptionsField
from .x_ms_odata import XmsODataField
from .x_ms_pageable import XmsPageableField

logger = logging.getLogger('aaz')


class Operation(Model, Linkable):
    """Describes a single API operation on a path."""

    tags = ListType(StringType())  # A list of tags for API documentation control. Tags can be used for logical grouping of operations by resources or any other qualifier.
    summary = StringType()  # A short summary of what the operation does. For maximum readability in the swagger-ui, this field SHOULD be less than 120 characters.
    description = StringType()  # A verbose explanation of the operation behavior. GFM syntax can be used for rich text representation.
    external_docs = ModelType(
        ExternalDocumentation,
        serialized_name="externalDocs",
        deserialize_from="externalDocs"
    )  # TODO: # Additional external documentation for this operation.
    operation_id = StringType(
        serialized_name="operationId",
        deserialize_from="operationId",
    )  # Unique string used to identify the operation. The id MUST be unique among all operations described in the API. Tools and libraries MAY use the operationId to uniquely identify an operation, therefore, it is recommended to follow common programming naming conventions.
    consumes = ListType(MimeField())  # TODO: # A list of MIME types the operation can consume. This overrides the consumes definition at the Swagger Object. An empty value MAY be used to clear the global definition. Value MUST be as described under Mime Types.
    produces = ListType(MimeField())  # TODO: # A list of MIME types the operation can produce. This overrides the produces definition at the Swagger Object. An empty value MAY be used to clear the global definition. Value MUST be as described under Mime Types.
    parameters = ListType(ParameterField(support_reference=True))  # A list of parameters that are applicable for this operation. If a parameter is already defined at the Path Item, the new definition will override it, but can never remove it. The list MUST NOT include duplicated parameters. A unique parameter is defined by a combination of a name and location. The list can use the Reference Object to link to parameters that are defined at the Swagger Object's parameters. There can be one "body" parameter at most.
    responses = DictType(PolyModelType([
        Reference, Response
    ]), required=True)  # The list of possible responses as they are returned from executing this operation.
    schemes = ListType(StringType(
        choices=("http", "https", "ws", "wss")
    ))  # TODO: # The transfer protocol for the operation. Values MUST be from the list: "http", "https", "ws", "wss". The value overrides the Swagger Object schemes definition.
    deprecated = BooleanType(default=False)  # TODO: # Declares this operation to be deprecated. Usage of the declared operation should be refrained. Default value is false.
    security = ListType(SecurityRequirementField())  # TODO: # A declaration of which security schemes are applied for this operation. The list of values describes alternative security schemes that can be used (that is, there is a logical OR between the security requirements). This definition overrides any declared top-level security. To remove a top-level security declaration, an empty array can be used.

    x_ms_pageable = XmsPageableField()  # TODO:
    x_ms_long_running_operation = XmsLongRunningOperationField(default=False)
    x_ms_long_running_operation_options = XmsLongRunningOperationOptionsField()

    x_ms_odata = XmsODataField()  # TODO: # indicates the operation includes one or more OData query parameters.
    x_ms_request_id = XmsRequestIdField()
    x_ms_examples = XmsExamplesField()

    # specific properties
    _x_publish = XPublishField()  # only used in Maps Data Plane
    _x_sf_codegen = XSfCodeGenField()  # only used in ServiceFabricMesh Mgmt Plane
    _x_ms_client_name = XmsClientNameField()  # only used in Maps Data Plane

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.x_ms_odata_instance = None
        self.x_ms_lro_final_state_schema = None

    def link(self, swagger_loader, *traces):
        if self.is_linked():
            return
        super().link(swagger_loader, *traces)

        if self.parameters is not None:
            for idx, param in enumerate(self.parameters):
                if isinstance(param, Linkable):
                    param.link(swagger_loader, *self.traces, 'parameters', idx)

            # replace parameter reference by parameter instance
            for idx in range(len(self.parameters)):
                param = self.parameters[idx]
                while isinstance(param, Reference):
                    param = param.ref_instance
                assert isinstance(param, ParameterBase)
                self.parameters[idx] = param

        # current response

        # verify path is resource id template or not
        resource_id_template = None
        if self.traces[-1] == 'get':
            # path should support get method
            resource_path = self.traces[-2]
            resource_id_template = swagger_resource_path_to_resource_id_template(resource_path)

        for key, response in self.responses.items():
            if resource_id_template and key != "default" and int(key) < 300:
                response.link(
                    swagger_loader, *self.traces, 'responses', key,
                    resource_id_template=resource_id_template
                )
            else:
                response.link(
                    swagger_loader, *self.traces, 'responses', key
                )

        # replace response reference by response instance
        for key in [*self.responses.keys()]:
            resp = self.responses[key]
            while isinstance(resp, Reference):
                if resp.ref_instance is None:
                    raise exceptions.InvalidSwaggerValueError(
                        msg="Reference not exist",
                        key=[],
                        value=resp.ref
                    )
                resp = resp.ref_instance
            assert isinstance(resp, Response)
            self.responses[key] = resp

        if self.x_ms_odata is not None:
            self.x_ms_odata_instance, instance_traces = swagger_loader.load_ref(
                self.x_ms_odata, *self.traces, 'x_ms_odata')
            if isinstance(self.x_ms_odata_instance, Linkable):
                self.x_ms_odata_instance.link(swagger_loader, *instance_traces)

        if self.x_ms_long_running_operation_options is not None and \
                self.x_ms_long_running_operation_options.final_state_schema is not None:
            # `final-state-schema` to `$ref`
            self.x_ms_lro_final_state_schema = ReferenceSchema()
            self.x_ms_lro_final_state_schema.ref = self.x_ms_long_running_operation_options.final_state_schema
            self.x_ms_lro_final_state_schema.link(
                swagger_loader,
                *self.traces, "x_ms_long_running_operation_options", "final_state_schema"
            )

        if self.x_ms_examples is not None:
            for key, example in self.x_ms_examples.items():
                try:
                    example.link(swagger_loader, *self.traces, "x_ms_examples", key)
                except Exception as e:
                    logger.error(f"Link example failed: {e}: {key}.")

    def to_cmd(self, builder, parent_parameters, host_path, **kwargs):
        cmd_op = CMDHttpOperation()
        if self.x_ms_long_running_operation:
            cmd_op.long_running = CMDHttpOperationLongRunning()
            if self.x_ms_long_running_operation_options:
                cmd_op.long_running.final_state_via = self.x_ms_long_running_operation_options.final_state_via

        cmd_op.operation_id = self.operation_id
        builder.setup_description(cmd_op, self)

        cmd_op.http = CMDHttpAction()
        if host_path:
            cmd_op.http.path = host_path + builder.path
        else:
            cmd_op.http.path = builder.path
        cmd_op.http.request = CMDHttpRequest()
        cmd_op.http.request.method = builder.method
        cmd_op.http.responses = []

        # request
        request = cmd_op.http.request

        param_models = {}
        client_request_id_name = None
        if parent_parameters:
            for p in parent_parameters:
                model = builder(p)
                if model is None:
                    continue
                if p.IN_VALUE not in param_models:
                    param_models[p.IN_VALUE] = {}
                param_models[p.IN_VALUE][p.name] = model
                if p.IN_VALUE == HeaderParameter.IN_VALUE:
                    if p.x_ms_client_request_id or p.name == "x-ms-client-request-id":
                        client_request_id_name = p.name

        if self.parameters:
            for p in self.parameters:
                model = builder(p)
                if model is None:
                    continue
                if p.IN_VALUE not in param_models:
                    param_models[p.IN_VALUE] = {}
                param_models[p.IN_VALUE][p.name] = model
                if p.IN_VALUE == HeaderParameter.IN_VALUE:
                    if p.x_ms_client_request_id or p.name == "x-ms-client-request-id":
                        client_request_id_name = p.name

        if PathParameter.IN_VALUE in param_models:
            request.path = CMDHttpRequestPath()
            request.path.params = []
            for _, model in sorted(param_models[PathParameter.IN_VALUE].items()):
                request.path.params.append(model)

        if QueryParameter.IN_VALUE in param_models:
            request.query = CMDHttpRequestQuery()
            request.query.params = []
            for _, model in sorted(param_models[QueryParameter.IN_VALUE].items()):
                request.query.params.append(model)

        if HeaderParameter.IN_VALUE in param_models:
            request.header = CMDHttpRequestHeader()
            request.header.client_request_id = client_request_id_name
            request.header.params = []
            for name, model in sorted(param_models[HeaderParameter.IN_VALUE].items()):
                if name == client_request_id_name:
                    # ignore client request id parameter for argument generation.
                    continue
                request.header.params.append(model)

        if BodyParameter.IN_VALUE in param_models:
            if len(param_models[BodyParameter.IN_VALUE]) > 1:
                raise exceptions.InvalidSwaggerValueError(
                    msg="Duplicate parameters in request body",
                    key=self.traces,
                    value=[builder.path, builder.method, builder.mutability, *param_models[BodyParameter.IN_VALUE].keys()]
                )
            model = [*param_models[BodyParameter.IN_VALUE].values()][0]
            if isinstance(model, CMDRequestJson):
                request.body = CMDHttpRequestJsonBody()
                request.body.json = model
            else:
                raise NotImplementedError()

        if FormDataParameter.IN_VALUE in param_models:
            raise NotImplementedError()

        # response
        success_responses,  redirect_responses, error_responses = builder.classify_responses(self)

        cmd_op.http.responses = []
        for status_codes, resp in success_responses:
            model = builder(resp, status_codes=status_codes)
            cmd_op.http.responses.append(model)
        for status_codes, resp in redirect_responses:
            model = builder(resp, status_codes=status_codes)
            cmd_op.http.responses.append(model)
        for status_codes, resp in error_responses:
            model = builder(resp, is_error=True, status_codes=status_codes)
            cmd_op.http.responses.append(model)

        if len(cmd_op.http.responses) == 0:
            raise exceptions.InvalidSwaggerValueError(
                msg="No http response",
                key=self.traces,
                value=[builder.path, builder.method]
            )
        return cmd_op



from utils import exceptions

from ..model.configuration import (CMDArg, CMDBooleanArg, CMDClsArg, CMDArrayArg, CMDArgumentHelp, CMDArgDefault,
                                   CMDPasswordArgPromptInput, CMDArgPromptInput, CMDFloatArgBase, CMDIntegerArgBase,
                                   CMDStringArgBase)


class ArgumentUpdateMixin:

    @classmethod
    def _update_arg(cls, arg, **kwargs):
        if isinstance(arg, CMDArg):
            cls._update_cmd_arg(arg, **kwargs)
            cls._update_arg_enum(arg, **kwargs)
        if isinstance(arg, CMDBooleanArg):
            cls._update_boolean_arg(arg, **kwargs)
        if isinstance(arg, CMDClsArg):
            cls._update_cls_arg(arg, **kwargs)
        if isinstance(arg, CMDArrayArg):
            cls._update_array_arg(arg, **kwargs)
            cls._update_arg_enum(arg.item, **kwargs)
        return arg

    @staticmethod
    def _update_cmd_arg(arg, **kwargs):
        if 'options' in kwargs:
            arg.options = kwargs['options']
        if 'stage' in kwargs:
            arg.stage = kwargs['stage']
        if 'hide' in kwargs:
            if kwargs['hide'] and arg.required:
                raise exceptions.ResourceConflict("Cannot hide required argument")
            arg.hide = kwargs['hide']
        if 'group' in kwargs:
            arg.group = kwargs['group']
        if 'help' in kwargs:
            arg.help = CMDArgumentHelp(kwargs['help'])
        if 'default' in kwargs:
            if kwargs['default'] is None:
                arg.default = None
            else:
                arg.default = CMDArgDefault(kwargs['default'])
        if 'configurationKey' in kwargs:
            arg.configuration_key = kwargs['configurationKey']
        if 'prompt' in kwargs:
            if kwargs['prompt'] is None:
                arg.prompt = None
            elif 'confirm' in kwargs['prompt']:
                arg.prompt = CMDPasswordArgPromptInput(kwargs['prompt'])
                arg.blank = None
            else:
                arg.prompt = CMDArgPromptInput(kwargs['prompt'])
                arg.blank = None

    @staticmethod
    def _update_boolean_arg(arg, **kwargs):
        if 'reverse' in kwargs:
            arg.reverse = kwargs['reverse'] or False

    @staticmethod
    def _update_cls_arg(arg, **kwargs):
        if 'singularOptions' in kwargs:
            arg.singular_options = kwargs['singularOptions'] or None

    @staticmethod
    def _update_array_arg(arg, **kwargs):
        if 'singularOptions' in kwargs:
            arg.singular_options = kwargs['singularOptions'] or None

    @staticmethod
    def _update_arg_enum(arg, **kwargs):
        if 'supportEnumExtension' in kwargs and isinstance(arg, (CMDFloatArgBase, CMDIntegerArgBase, CMDStringArgBase)):
            if arg.enum:
                arg.enum.support_extension = kwargs['supportEnumExtension']


import json
import logging
import os
import re
import shutil

from command.model.configuration import CMDConfiguration, CMDHelp, CMDCommandExample, XMLSerializer, CMDClientConfig
from utils.base64 import b64encode_str
from utils.config import Config
from utils.plane import PlaneEnum
from command.model.specs import CMDSpecsCommandTree, CMDSpecsCommandGroup, CMDSpecsCommand, CMDSpecsCommandVersion, CMDSpecsResource
from command.templates import get_templates
from utils import exceptions
from .cfg_reader import CfgReader
from .client_cfg_reader import ClientCfgReader
from .cfg_validator import CfgValidator
from .command_tree import CMDSpecsPartialCommandTree

logger = logging.getLogger('aaz')


class AAZSpecsManager:
    COMMAND_TREE_ROOT_NAME = "aaz"

    REFERENCE_LINE = re.compile(r"^Reference\s*\[(.*) (.*)]\((.*)\)\s*$")

    def __init__(self):
        if not Config.AAZ_PATH or not os.path.exists(Config.AAZ_PATH) or not os.path.isdir(Config.AAZ_PATH):
            raise ValueError(f"aaz repo path is invalid: '{Config.AAZ_PATH}'")

        self.folder = Config.AAZ_PATH
        self.resources_folder = os.path.join(self.folder, "Resources")
        self.commands_folder = os.path.join(self.folder, "Commands")
        self._tree = None
        self._modified_resource_cfgs = {}
        self._modified_resource_client_cfgs = {}

        self._tree = CMDSpecsPartialCommandTree(self.folder)

    @property
    def tree(self):
        return self._tree

    @property
    def simple_tree(self):
        return self.tree.simple_tree

    # Commands folder
    def get_tree_file_path(self):
        return os.path.join(self.commands_folder, "tree.json")

    def get_command_group_folder(self, *cg_names):
        # support len(cg_names) == 0
        return os.path.join(self.commands_folder, *cg_names)

    def get_command_group_readme_path(self, *cg_names):
        # when len(cg_names) == 0, the path will be the CommandGroup/readme.md
        return os.path.join(self.get_command_group_folder(*cg_names), "readme.md")

    def get_command_readme_path(self, *cmd_names):
        if len(cmd_names) <= 1:
            raise ValueError(f"Invalid command names: '{' '.join(cmd_names)}'")
        return os.path.join(self.get_command_group_folder(*cmd_names[:-1]), f"_{cmd_names[-1]}.md")

    # resources
    def get_resource_plane_folder(self, plane):
        if plane == PlaneEnum.Mgmt:
            return os.path.join(self.resources_folder, plane)
        elif PlaneEnum.is_data_plane(plane):
            scope = PlaneEnum.get_data_plane_scope(plane)
            if not scope:
                raise ValueError(f"Invalid plane: Missing scope in data plane '{plane}'")
            return os.path.join(self.resources_folder, PlaneEnum._Data, scope)
        else:
            raise ValueError(f"Invalid plane: '{plane}'")
    
    def get_resource_client_cfg_paths(self, plane):
        path = self.get_resource_plane_folder(plane)
        return os.path.join(path, "client.json"), os.path.join(path, "client.xml")

    def get_resource_cfg_folder(self, plane, resource_id):
        path = self.get_resource_plane_folder(plane)
        name = b64encode_str(resource_id)
        while len(name):
            if len(name) > 255:
                path = os.path.join(path, name[:254] + '+')
                name = name[254:]
            else:
                path = os.path.join(path, name)
                name = ""
        return path

    def get_resource_cfg_file_paths(self, plane, resource_id, version):
        """Return Json and XML path"""
        path = os.path.join(self.get_resource_cfg_folder(plane, resource_id), f"{version}")
        return f"{path}.json", f"{path}.xml"

    def get_resource_cfg_ref_file_path(self, plane, resource_id, version):
        return os.path.join(self.get_resource_cfg_folder(plane, resource_id), f"{version}.md")

    def get_resource_versions(self, plane, resource_id):
        path = self.get_resource_cfg_folder(plane, resource_id)
        if not os.path.exists(path) or not os.path.isdir(path):
            return None
        versions = set()
        for file_name in os.listdir(path):
            if file_name.endswith('.xml'):
                versions.add(file_name[:-4])
            elif file_name.endswith('.json'):
                versions.add(file_name[:-5])
            elif file_name.endswith('.md'):
                versions.add(file_name[:-3])
        return sorted(versions, reverse=True)

    def find_command_group(self, *cg_names):
        return self.tree.find_command_group(*cg_names)

    def find_command(self, *cmd_names):
        return self.tree.find_command(*cmd_names)

    def iter_command_groups(self, *root_cg_names):
        yield from self.tree.iter_command_groups(*root_cg_names)

    def iter_commands(self, *root_node_names):
        yield from self.tree.iter_commands(*root_node_names)

    def load_resource_cfg_reader(self, plane, resource_id, version):
        key = (plane, resource_id, version)
        if key in self._modified_resource_cfgs:
            # cfg already modified
            cfg = self._modified_resource_cfgs[key]
            return CfgReader(cfg) if cfg else None

        json_path, xml_path = self.get_resource_cfg_file_paths(plane, resource_id, version)
        if not os.path.exists(json_path) and not os.path.exists(xml_path):
            ref_path = self.get_resource_cfg_ref_file_path(plane, resource_id, version)
            if not os.path.exists(ref_path):
                return None
            json_path = None
            with open(ref_path, 'r', encoding="utf-8") as f:
                for line in f.readlines():
                    match = self.REFERENCE_LINE.fullmatch(line)
                    if match:
                        resource_id, version = match[1], match[2]
                        json_path, xml_path = self.get_resource_cfg_file_paths(plane, resource_id, version)
                        break
            if not json_path or not os.path.exists(json_path):
                raise ValueError(f"Invalid reference file: {ref_path}")

        if not os.path.exists(json_path) and os.path.exists(xml_path):
            if not os.path.isfile(xml_path):
                raise ValueError(f"Invalid file path: {xml_path}")
            # Convert existing xml to json.
            # Not recommend to use xml, because there are some issues in XMLSerializer
            if not os.path.isfile(xml_path):
                raise ValueError(f"Invalid file path: {xml_path}")
            with open(xml_path, 'r', encoding="utf-8") as f:
                cfg = XMLSerializer.from_xml(CMDConfiguration, f.read())
            data = self.render_resource_cfg_to_json(cfg)
            with open(json_path, 'w', encoding="utf-8") as f:
                f.write(data)
            data = self.render_resource_cfg_to_xml(cfg)
            with open(xml_path, 'w', encoding="utf-8") as f:
                f.write(data)

        if not os.path.isfile(json_path):
            raise ValueError(f"Invalid file path: {json_path}")

        with open(json_path, 'r', encoding="utf-8") as f:
            #print(json_path)
            data = json.load(f)
        cfg = CMDConfiguration(data)

        return CfgReader(cfg)

    def load_resource_cfg_reader_by_command_with_version(self, cmd, version):
        if not isinstance(version, CMDSpecsCommandVersion):
            assert isinstance(version, str)
            version_name = version
            version = None
            for v in cmd.versions or []:
                if v.name == version_name:
                    version = v
                    break
        if not version:
            return None
        resource = version.resources[0]
        return self.load_resource_cfg_reader(resource.plane, resource.id, resource.version)

    # command tree
    def create_command_group(self, *cg_names):
        return self.tree.create_command_group(*cg_names)

    def update_command_group_by_ws(self, ws_node):
        return self.tree.update_command_group_by_ws(ws_node)

    def delete_command_group(self, *cg_names):
        return self.tree.delete_command_group(*cg_names)
    
    def create_command(self, *cmd_names):
        return self.tree.create_command(*cmd_names)
    
    def delete_command(self, *cmd_names):
        return self.tree.delete_command(*cmd_names)
    
    def delete_command_version(self, *cmd_names, version):
        return self.tree.delete_command_version(*cmd_names, version=version)
    
    def update_command_version(self, *cmd_names, plane, cfg_cmd):
        return self.tree.update_command_version(*cmd_names, plane=plane, cfg_cmd=cfg_cmd)
    
    def update_command_by_ws(self, ws_leaf):
        return self.tree.update_command_by_ws(ws_leaf)
    
    def verify_command_tree(self):
        return self.tree.verify_command_tree()

    def verify_updated_command_tree(self):
        return self.tree.verify_updated_command_tree()

    def _remove_cfg(self, cfg):
        cfg_reader = CfgReader(cfg)

        # update resource cfg
        for resource in cfg_reader.resources:
            key = (cfg.plane, resource.id, resource.version)
            self._modified_resource_cfgs[key] = None

        # update command tree
        for cmd_names, cmd in cfg_reader.iter_commands():
            self.delete_command_version(*cmd_names, version=cmd.version)

    def update_resource_cfg(self, cfg):
        cfg_reader = CfgReader(cfg=cfg)

        cfg_verifier = CfgValidator(cfg_reader)
        # TODO: implement verify configuration
        cfg_verifier.verify()

        # remove previous cfg
        for resource in cfg_reader.resources:
            pre_cfg_reader = self.load_resource_cfg_reader(cfg.plane, resource_id=resource.id, version=resource.version)
            if pre_cfg_reader and pre_cfg_reader.cfg != cfg:
                self._remove_cfg(pre_cfg_reader.cfg)

        # add new command version
        for cmd_names, cmd in cfg_reader.iter_commands():
            self.update_command_version(*cmd_names, plane=cfg.plane, cfg_cmd=cmd)

        for resource in cfg_reader.resources:
            key = (cfg.plane, resource.id, resource.version)
            self._modified_resource_cfgs[key] = cfg

    # client configuration
    def load_client_cfg_reader(self, plane):
        key = (plane, )
        if key in self._modified_resource_client_cfgs:
            # cfg already modified
            cfg = self._modified_resource_client_cfgs[key]
            return ClientCfgReader(cfg) if cfg else None

        json_path, xml_path = self.get_resource_client_cfg_paths(plane)
        if not os.path.exists(json_path) and not os.path.exists(xml_path):
            return None
        
        if not os.path.exists(json_path) and os.path.exists(xml_path):
            if not os.path.isfile(xml_path):
                raise ValueError(f"Invalid file path: {xml_path}")
            # Convert existing xml to json.
            # Not recommend to use xml, because there are some issues in XMLSerializer
            if not os.path.isfile(xml_path):
                raise ValueError(f"Invalid file path: {xml_path}")
            with open(xml_path, 'r') as f:
                cfg = XMLSerializer.from_xml(CMDClientConfig, f.read())
            data = self.render_resource_cfg_to_json(cfg)
            with open(json_path, 'w') as f:
                f.write(data)
            data = self.render_resource_cfg_to_xml(cfg)
            with open(xml_path, 'w') as f:
                f.write(data)
        
        if not os.path.isfile(json_path):
            raise ValueError(f"Invalid file path: {json_path}")
        
        with open(json_path, 'r') as f:
            data = json.load(f)
        cfg = CMDClientConfig(data)
        return ClientCfgReader(cfg)

    def update_client_cfg(self, cfg):
        """This function guarantee the client config version is always increasing."""
        assert isinstance(cfg, CMDClientConfig)
        if old_cfg_reader := self.load_client_cfg_reader(cfg.plane):
            if old_cfg_reader.cfg.version > cfg.version:
                raise exceptions.InvalidAPIUsage("Failed to update: a new version of client config exists.")
            if cfg.version == old_cfg_reader.cfg.version:
                # didn't change when version is same
                return
        key = (cfg.plane, )
        self._modified_resource_client_cfgs[key] = cfg

    def save(self):
        self.verify_updated_command_tree()

        remove_files = []
        remove_folders = []
        update_files = {}
        command_groups = set()

        # command
        for cmd_names in sorted(self.tree._modified_commands):
            cmd = self.find_command(*cmd_names)
            file_path = self.get_command_readme_path(*cmd_names)
            if not cmd:
                # remove command file
                remove_files.append(file_path)
            else:
                update_files[file_path] = self.render_command_readme(cmd)

            command_groups.add(tuple(cmd_names[:-1]))

        for cg_names in sorted(self.tree._modified_command_groups):
            command_groups.add(tuple(cg_names))
            command_groups.add(tuple(cg_names[:-1]))

        # command groups
        for cg_names in sorted(command_groups):
            cg = self.find_command_group(*cg_names)
            if not cg:
                # remove command group folder
                remove_folders.append(self.get_command_group_folder(*cg_names))
            else:
                # update command group readme
                file_path = self.get_command_group_readme_path(*cg_names)
                if cg == self.tree.root:
                    update_files[file_path] = self.render_command_tree_readme(self.tree)
                else:
                    update_files[file_path] = self.render_command_group_readme(cg)

        # cfg files
        for (plane, resource_id, version), cfg in self._modified_resource_cfgs.items():
            json_file_path, xml_file_path = self.get_resource_cfg_file_paths(plane, resource_id, version)
            ref_file_path = self.get_resource_cfg_ref_file_path(plane, resource_id, version)
            if not cfg:
                remove_files.append(json_file_path)
                remove_files.append(xml_file_path)
                remove_files.append(ref_file_path)
            else:
                main_resource = cfg.resources[0]
                if main_resource.id != resource_id or main_resource.version != version:
                    update_files[ref_file_path] = self.render_resource_ref_readme(
                        plane=cfg.plane, ref_resource_id=main_resource.id, ref_resource_version=main_resource.version)
                else:
                    update_files[json_file_path] = self.render_resource_cfg_to_json(cfg)
                    update_files[xml_file_path] = self.render_resource_cfg_to_xml(cfg)
        
        # client cfg files
        for (plane, ), cfg in self._modified_resource_client_cfgs.items():
            json_file_path, xml_file_path = self.get_resource_client_cfg_paths(plane)
            assert isinstance(cfg, CMDClientConfig)
            update_files[json_file_path] = self.render_resource_cfg_to_json(cfg)
            update_files[xml_file_path] = self.render_resource_cfg_to_xml(cfg)

        for remove_file in remove_files:
            if os.path.exists(remove_file):
                os.remove(remove_file)

        for remove_folder in remove_folders:
            shutil.rmtree(remove_folder, ignore_errors=True)

        for file_path, data in update_files.items():
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(file_path, 'w', encoding="utf-8") as f:
                f.write(data)

        self._modified_resource_cfgs = {}
        self._modified_resource_client_cfgs = {}

    @staticmethod
    def render_command_readme(command):
        assert isinstance(command, CMDSpecsCommand)
        tmpl = get_templates()['command']
        return tmpl.render(command=command)

    @staticmethod
    def render_command_group_readme(command_group):
        assert isinstance(command_group, CMDSpecsCommandGroup)
        tmpl = get_templates()['group']
        return tmpl.render(group=command_group)

    @staticmethod
    def render_command_tree_readme(tree):
        assert isinstance(tree, (CMDSpecsCommandTree, CMDSpecsPartialCommandTree))
        tmpl = get_templates()['tree']
        return tmpl.render(tree=tree)

    @staticmethod
    def render_resource_ref_readme(plane, ref_resource_id, ref_resource_version):
        ref_resource = CMDSpecsResource({
            "plane": plane,
            "id": ref_resource_id,
            "version": ref_resource_version
        })
        tmpl = get_templates()['resource_ref']
        return tmpl.render(ref_resource=ref_resource)

    @staticmethod
    def render_resource_cfg_to_json(cfg):
        data = cfg.to_primitive()
        return json.dumps(data, ensure_ascii=False)

    @staticmethod
    def render_resource_cfg_to_xml(cfg):
        return XMLSerializer.to_xml(cfg)



from schematics.types import StringType, BaseType, BooleanType
from utils.stage import AAZStageEnum, AAZStageField
import json
import logging


logger = logging.getLogger('aaz')


class CMDBooleanField(BooleanType):

    def __init__(self, **kwargs):
        super(CMDBooleanField, self).__init__(serialize_when_none=False, default=False, **kwargs)

    def to_native(self, value, context=None):
        value = super(CMDBooleanField, self).to_native(value, context)
        if value is False:
            return None  # return None when value is false to hide field with `serialize_when_none=False`
        return value

    def to_primitive(self, value, context=None):
        value = super(CMDBooleanField, self).to_primitive(value, context)
        if value is False:
            return None  # return None when value is false to hide field with `serialize_when_none=False`
        return value


class CMDStageField(AAZStageField):
    """The stage for command group, command or argument."""

    def to_native(self, value, context=None):
        value = super(CMDStageField, self).to_native(value, context)
        if value == AAZStageEnum.Stable:
            return None  # return None when value is false to hide field with `serialize_when_none=False`
        return value

    def to_primitive(self, value, context=None):
        value = super(CMDStageField, self).to_primitive(value, context)
        if value == AAZStageEnum.Stable:
            return None  # return None when value is false to hide field with `serialize_when_none=False`
        return value


class CMDVariantField(StringType):
    """The variant definition"""

    def __init__(self, *args, **kwargs):
        super(CMDVariantField, self).__init__(
            regex=r'[$@][a-zA-Z0-9_\[\]\{\}\.]+',
            *args,
            **kwargs
        )


class CMDClassField(StringType):

    def __init__(self, *args, **kwargs):
        super(CMDClassField, self).__init__(
            regex=r'[A-Z][a-zA-Z0-9_]+',
            *args, **kwargs
        )


class CMDPrimitiveField(BaseType):
    """
    Can parse json value format string, the result type can be null, integer, float, bool, string, list or dict
    """

    @staticmethod
    def convert_from_xml(raw_data):
        try:
            return json.loads(raw_data)
        except json.JSONDecodeError as err:
            logging.error(f'Parse primitive field value from xml failed: "{raw_data}"({type(raw_data)}): {err}\n'
                          f'Please export Command Model field again. ')
            return raw_data

    def __init__(self, *args, **kwargs):
        super(CMDPrimitiveField, self).__init__(
            serialize_when_none=True,
            default=None,
            *args, **kwargs,
        )

    def to_primitive(self, value, context=None):
        # TODO: when value is None, will not call to_primitive, so it will not be converted to null in json.
        data = super().to_primitive(value, context)
        if context is not None and "to_xml" in context and context.to_xml is True:
            data = json.dumps(data, ensure_ascii=False, indent=None, separators=(',', ':'))
        return data


class CMDRegularExpressionField(StringType):
    # This string SHOULD be a valid regular expression
    pass


class CMDVersionField(StringType):

    def __init__(self, *args, **kwargs):
        super(CMDVersionField, self).__init__(*args, **kwargs)


class CMDConfirmation(StringType):

    def __int__(self, *args, **kwargs):
        super(CMDConfirmation, self).__init__(*args, **kwargs)


class CMDResourceIdField(StringType):

    def __init__(self, *args, **kwargs):
        super(CMDResourceIdField, self).__init__(
            serialized_name='id',
            deserialize_from='id',
            *args, **kwargs
        )


class CMDCommandNameField(StringType):

    def __init__(self, *args, **kwargs):
        super(CMDCommandNameField, self).__init__(
            regex=r'^[a-z0-9]+(-[a-z0-9]+)*$',
            min_length=1, *args, **kwargs)


class CMDCommandGroupNameField(StringType):

    def __init__(self, *args, **kwargs):
        super(CMDCommandGroupNameField, self).__init__(
            regex=r'^[a-z0-9]+(-[a-z0-9]+)*( [a-z0-9]+(-[a-z0-9]+)*)*$',
            min_length=1, *args, **kwargs)


class CMDURLPathField(StringType):

    def __init__(self, *args, **kwargs):
        super(CMDURLPathField, self).__init__(*args, **kwargs)


class CMDDescriptionField(StringType):
    """The description information from swagger. It's the source of helps. Should not be saved in configuration file."""

    def __init__(self, *args, **kwargs):
        super(CMDDescriptionField, self).__init__(
            serialize_when_none=False,
            *args,
            **kwargs
        )

    def to_primitive(self, value, context=None):
        """the description will not exist when call to primitive"""
        return None  # return None when value is false to hide field with `serialize_when_none=False`




_templates = None


def get_templates():
    global _templates
    if _templates is None:
        import os
        from jinja2 import Environment, FileSystemLoader
        from ._filters import custom_filters
        env = Environment(loader=FileSystemLoader(searchpath=os.path.dirname(os.path.abspath(__file__))))
        env.filters.update(custom_filters)
        _templates = {}
        _templates['tree'] = env.get_template("tree.md.j2")
        _templates['group'] = env.get_template("group.md.j2")
        _templates['command'] = env.get_template("command.md.j2")
        _templates['resource_ref'] = env.get_template("resource_ref.md.j2")
    return _templates



from schematics.models import Model
from schematics.types import PolyModelType

from ._arg_builder import CMDArgBuilder
from ._fields import CMDVariantField
from ._schema import CMDSchemaBaseField, CMDSchema, CMDClsSchema, CMDClsSchemaBase, \
    CMDObjectSchemaBase, CMDArraySchemaBase, CMDObjectSchemaDiscriminator
from ._utils import CMDDiffLevelEnum


class CMDRequestJson(Model):
    """Used for Request Body and Instance Update operation"""

    # properties as tags
    ref = CMDVariantField()

    # properties as nodes
    schema = PolyModelType(CMDSchema, allow_subclasses=True)

    class Options:
        serialize_when_none = False

    def generate_args(self, ref_args, var_prefix=None, is_update_action=False):
        if not self.schema:
            return []
        assert isinstance(self.schema, CMDSchema)
        builder = CMDArgBuilder.new_builder(
            schema=self.schema,
            ref_args=ref_args,
            var_prefix=var_prefix,
            is_update_action=is_update_action
        )
        args = builder.get_args()
        return args

    def diff(self, old, level):
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            if (self.ref is not None) != (old.ref is not None):
                diff["ref"] = f"{old.ref} != {self.ref}"
            schema_diff = self.schema.diff(old.schema, level)
            if schema_diff:
                diff["schema"] = schema_diff

        if level >= CMDDiffLevelEnum.Associate:
            if self.ref != old.ref:
                diff["ref"] = f"{old.ref} != {self.ref}"
        return diff

    def reformat(self, schema_cls_map, **kwargs):
        if self.schema:
            if getattr(self.schema, 'cls', None):
                if not schema_cls_map.get(self.schema.cls, None):
                    schema_cls_map[self.schema.cls] = self.schema
                else:
                    # replace by CMDClsSchema
                    self.schema = CMDClsSchema.build_from_schema(self.schema, schema_cls_map[self.schema.cls])

            _iter_over_schema(self.schema, schema_cls_map)

            self.schema.reformat(**kwargs)
    
    def register_cls(self, cls_register_map, **kwargs):
        _iter_over_schema_for_cls_register(self.schema, cls_register_map)


class CMDResponseJson(Model):
    # properties as tags
    var = CMDVariantField()

    # properties as nodes
    schema = CMDSchemaBaseField(required=True)

    class Options:
        serialize_when_none = False

    def diff(self, old, level):
        diff = {}
        if level >= CMDDiffLevelEnum.BreakingChange:
            schema_diff = self.schema.diff(old.schema, level)
            if schema_diff:
                diff["schema"] = schema_diff

        if level >= CMDDiffLevelEnum.Associate:
            if self.var != old.var:
                diff["var"] = f"{old.var} != {self.var}"
        return diff

    def reformat(self, schema_cls_map, **kwargs):
        if getattr(self.schema, 'cls', None):
            if self.schema.cls in schema_cls_map:
                self.schema = CMDClsSchemaBase.build_from_schema_base(self.schema, schema_cls_map[self.schema.cls])
            else:
                schema_cls_map[self.schema.cls] = self.schema
        _iter_over_schema(self.schema, schema_cls_map)
        self.schema.reformat(**kwargs)
    
    def register_cls(self, cls_register_map, **kwargs):
        _iter_over_schema_for_cls_register(self.schema, cls_register_map)


def _iter_over_schema(schema, schema_cls_map):
    if schema.frozen:
        return
    if isinstance(schema, (CMDObjectSchemaBase, CMDObjectSchemaDiscriminator)):
        if schema.props:
            for idx in range(len(schema.props)):
                s = schema.props[idx]
                if getattr(s, 'cls', None):
                    if not schema_cls_map.get(s.cls, None):
                        schema_cls_map[s.cls] = s
                    else:
                        # replace by CMDClsSchema
                        schema.props[idx] = CMDClsSchema.build_from_schema(s, schema_cls_map[s.cls])
            for prop in schema.props:
                _iter_over_schema(prop, schema_cls_map)

        if schema.discriminators:
            for disc in schema.discriminators:
                _iter_over_schema(disc, schema_cls_map)

        if isinstance(schema, CMDObjectSchemaBase) and schema.additional_props and schema.additional_props.item:
            s = schema.additional_props.item
            if getattr(s, 'cls', None):
                if not schema_cls_map.get(s.cls, None):
                    schema_cls_map[s.cls] = s
                else:
                    # replace by CMDClsBaseSchema
                    schema.additional_props.item = CMDClsSchemaBase.build_from_schema_base(s, schema_cls_map[s.cls])
            _iter_over_schema(schema.additional_props.item, schema_cls_map)

    elif isinstance(schema, CMDArraySchemaBase):
        s = schema.item
        if getattr(s, 'cls', None):
            if not schema_cls_map.get(s.cls, None):
                schema_cls_map[s.cls] = s
            else:
                # replace by CMDClsBaseSchema
                schema.item = CMDClsSchemaBase.build_from_schema_base(s, schema_cls_map[s.cls])
        _iter_over_schema(schema.item, schema_cls_map)
    elif isinstance(schema, CMDClsSchemaBase):
        cls_name = schema.type[1:]
        if cls_name not in schema_cls_map:
            # set this cls name as None, in order to check where this cls_name miss cls definition
            schema_cls_map[cls_name] = None


def _iter_over_schema_for_cls_register(schema, cls_register_map):
    if schema is None or schema.frozen:
        return

    if isinstance(schema, CMDClsSchemaBase):
        cls_name = schema.type[1:]
        if cls_name not in cls_register_map:
            cls_register_map[cls_name] = {
                "implement": None,
                "refers": []
            }
        cls_register_map[cls_name]['refers'].append(schema)
        return 

    if getattr(schema, 'cls', None):
        cls_name = schema.cls
        if cls_name not in cls_register_map:
            cls_register_map[cls_name] = {
                "implement": None,
                "refers": []
            }
        cls_register_map[cls_name]['implement'] = schema

    if isinstance(schema, (CMDObjectSchemaBase, CMDObjectSchemaDiscriminator)):
        if schema.props:
            for prop in schema.props:
                _iter_over_schema_for_cls_register(prop, cls_register_map)
        
        if schema.discriminators:
            for disc in schema.discriminators:
                _iter_over_schema_for_cls_register(disc, cls_register_map)
        
        if isinstance(schema, CMDObjectSchemaBase) and schema.additional_props and schema.additional_props.item:
            _iter_over_schema_for_cls_register(schema.additional_props.item, cls_register_map)
    
    elif isinstance(schema, CMDArraySchemaBase):
        _iter_over_schema_for_cls_register(schema.item, cls_register_map)
